"""
info_array - define class and methods for the 'Information'-packed
ndarray. it inherits all the metods of the numpy array class
and adds the ability to set other attributes, pretty-print, and smart graphing.
"""
import os
import subprocess
import numpy as np
import dimg
plt = dimg.plt
import img
from units import units
import scipy.ndimage
import h5py
import pyfits

def copyattrs(sobj,dobj):
    d = sobj.__dict__
    keys = d.keys()
    for key in keys:
        setattr(dobj,key,d[key])
    
class InfoArray(np.ndarray):
    """
    A sub class of `numpy.ndarray <http://docs.scipy.org/doc/numpy/reference/arrays.ndarray.html>`_, with useful additional attributes like **name**, **dx**, and **units.**
    
    (See `numpy array subclassing <http://docs.scipy.org/doc/numpy-1.10.1/user/basics.subclassing.html>`_)
    
    to use::
    
        a = InfoArray(np.zeros((2,2)),name='blah')
        a.dx = ...
    
    InfoArrays are containers in that they can contain any number of attributes. Useful ones are:
    
    :ivar str name: The name of the array
    :ivar str units: The units of the values in the array
    :ivar float dx: The physical size of one pixel in the array
    :ivar str dx_units: The units of the pixel size
    :ivar list axis_names: A list of strings containing the axis names
        The first and second are the horizontal and vertical axes correspondingly.
        An optionall third is the radial (for radial average plots).
    :ivar list x0y0: Physical position of the [0,0] pixel (defaults to -shape*dx/2)
    
    """
    def __new__(cls, input_array, name='',**kwargs):
        obj = np.asarray(input_array).view(cls)
        if type(input_array) is InfoArray:
            copyattrs(input_array,obj)
            setattr(obj,'name',name)
                
        else:
            obj.name = name
            obj.units = ''
            obj.dx = 1.0
            obj.dx_units = ''
            obj.plotParams = {
                'fontsize.title':16,
                'fontsize.subtitle':13,
                'fontsize.axis':12,
                'fontsize.tick':10,
                'axislabel.left':True,
                'axislabel.right':True,
                'axislabel.top':True,
                'axislabel.bottom':True,
            }

        obj.__dict__.update(kwargs)
        return obj
    
    def __array_finalize__(self, obj):
        if obj is None: return
        if isinstance(obj,InfoArray):
            copyattrs(obj,self)
        
    def copy(self):
        """Return a deep copy of the InfoArray
        
        :rtype: InfoArray
        """
        obj = InfoArray(self.view(np.ndarray).copy())
        copyattrs(self,obj)
        return obj
    
    def minmax(self):
        """Calculate the miimum and maximum of the array
        
        :return: a list of [minimumm, maximum]

        """
        return (self.min(),self.max())    
    
    def pprint(self,full=False):
        """Pretty-print the object. This consists
        of reporting on the object's attributes.

        :param boolean full: if True, **pprint** also prints out the contents of the array.
            Default is to print only the "header" (attributes) information.

        """
        print self._pprint(full=full)
    
    def oft(self,upsamp=1):
        """Performs the Optical Fourier Transform on the array
        and returns a new InfoArray object. The object data are treated as
        optical path distances (OPD). An InfoArray object representing the aperture
        throughput can be included as an instance object, otherwise a default unit disk is used
        
        The forumla is
            PSF = F { ap exp {i (2 pi / wavelength) OPD } }
        
        OFT assumes that the object has the following instance variables, and if it
        does not, then default ones are set up:
            
            wavelength = wavelength of the light, in microns (default: 0.5)
            ap = an InfoArray object that defines the aperture (default: a circlular disk that assures Nyquist sampling)
            units = the units of the wavefront (default: 'meters')
            dx_units = the units of dx (default: 'meters')
        """
        
        if not hasattr(self,'wavelength'):
            self.wavelength = 0.5
            self.wavelength_units = 'microns'
        n,m = self.shape
        lam = self.wavelength*units[self.wavelength_units]
        k = 2.*np.pi/lam
        if not hasattr(self,'ap'):
            ap = img.circle((n,m),r=min(n,m)/4.)
            self.ap = InfoArray(ap,name='%s.ap'%self.name,
                                units = 'throughput',
                                dx = self.dx,
                                dx_units = self.dx_units,)
        dx = self.dx*units[self.dx_units]
        wf = self.ap*np.exp(1j*k*self*units[self.units])
        dth = lam/(n*dx)/units['arcsec']
        dth_units = 'arcsec'
        if upsamp in [2,4,8]:
            wf = img.zeropad(wf,(n*upsamp,m*upsamp))
            dth = upsamp*dth
        psf = np.abs( np.fft.ifftshift( np.fft.fft2( np.fft.fftshift(wf))))**2
        if upsamp != 1:
            psf = img.crop(psf,np.array(psf.shape)/2,(n,m))
        psf = InfoArray(psf/psf.max(),
                        name='OFT(%s)'%self.name,
                        units = 'normalized to peak',
                        dx = dth,
                        dx_units = dth_units,
                        wavelength = self.wavelength,
                        wavelength_units = self.wavelength_units)
        return psf
    
    def psf(self):
        """Compute the PSF (Point Spread Function) assuming the data are the MTF
        (Modulation Transfer Function). It is assumed that the MTF is symmetric
        so that the PSF is real
        """
        n,m = self.shape
        lam = self.wavelength*units[self.wavelength_units]
        k = 2.*np.pi/lam
        dx = self.dx*units[self.dx_units]
        dth = lam/(n*dx)/units['arcsec']
        dth_units = 'arcsec'
        psf = np.abs(np.fft.ifftshift(np.fft.fft2(np.fft.fftshift(self))))
        psf = InfoArray(psf/psf.max(),
                        name ='PSF(%s)'%self.name,
                        units = 'normalized to peak',
                        dx = dth,
                        dx_units = dth_units,
                        wavelength = self.wavelength,
                        wavelength_units = self.wavelength_units)
        return psf
    
    def mtf(self):
        """Compute the MTF (Modulation Transfer Function) assuming that the data
        are the PSF (Point Spread Function). It is assumed that the PSF is real
        """
        n,m = self.shape
        lam = self.wavelength*units[self.wavelength_units]
        k = 2.*np.pi/lam
        dx = self.dx*units[self.dx_units]
        dth = lam/(n*dx)
        dth_units = 'meters'
        mtf = np.abs(np.fft.ifftshift(np.fft.ifft2(np.fft.fftshift(self))))
        mtf = InfoArray(mtf/mtf.max(),
                        name ='MTF(%s)'%self.name,
                        units = 'normalized to peak',
                        dx = dth,
                        dx_units = dth_units,
                        wavelength = self.wavelength,
                        wavelength_units = self.wavelength_units)
        return mtf
    
    def calc_ena(self):
        """Calculate the Equivalent Noise Area, assuming the data represent
        a point spread function (PSF).
        Reference: King, PASP, v95, 1983
        
        Formula:
            ENA = dA / \sum_i f_i^2
            
        where dA is the area of a pixel (dx^2), and f is the psf normalized to integrate
        to unity.
        The PSF is assumed real.
        """
        f = self/self.sum()
        ena = self.dx**2/np.sum(f**2)
        ena_units = '%s^2'%self.dx_units
        ena_r = np.sqrt(ena/np.pi)
        ena_r_units = '%s'%self.dx_units
        self.ena,self.ena_units = (ena,ena_units)
        self.ena_r,self.ena_r_units = (ena_r,ena_r_units)
        return

    def resample(self,dx):
        """Resample the image so as to have a different dx.
        This will increase or decrease the dimensions of the data.        
        """
        r = scipy.ndimage.zoom(self,self.dx/dx)
        r = InfoArray(r)
        copyattrs(self,r)
        r.dx = dx
        r.name = r.name+'_zoomed'
        return r
    
    def resize(self,shape):
        """Resize the image by resampling. This will
        change the sampling dx. It is prevented from changing the
        aspect ratio.
        """
        n,m = self.shape
        nr,mr = shape
        fac = float(nr)/float(n)
        mrt = fac*float(m)
        if not np.isclose(mr,mrt,atol=1):
            raise Exception,'<resize> can\'t change aspect ratio'
        return self.resample(self.dx/fac)

    def crop(self,shape,p0=None):
        """Crop the image to a different size than the present image.
        This different size can be larger or smaller than the original image.
        p0 is an optional lower left hand corner for the crop
        """
        center = np.array(self.shape)/2
        if p0 is not None:
            center = np.array(p0) + np.array(shape)/2
        r = img.crop(self,center,shape)
        r = InfoArray(r)
        copyattrs(self,r)
        r.name = r.name+'_cropped'
        return r

    def _pprint(self,print_width=40,aslist=False,rshort=False,full=False):
        cr = '\n'
        bling = 50*'='
        rs = []
        rs.append(bling)
        if hasattr(self,'longName'):
            rs.append(self.longName)
        else:
            if hasattr(self,'name') and self.name is not None:
                rs.append(self.name)
        rs.append(str(type(self)))
        rs.append(bling)
        rs.append('shape: '+str(self.shape))
        if rshort:
            if aslist:
                return rs
            else:
                rs = '\n'.join(rs)
                return rs
        if full:
            rs.append(str(self))
        tab = ' '*4
        d = self.__dict__
        keys = sorted(d.keys())
        for key in keys:
            if key.endswith('_units'): continue
            val = d[key]
            units = None
            if isinstance(val,(str,unicode)): pv = "'"+val+"'"
            else: pv = str(val)
            if len(pv) > print_width:
                pv = str(val.__class__)
                if isinstance(val,(np.ndarray,np.matrix)):
                    pv += ' '+str(val.shape)
            if key+'_units' in keys:
                units = d[key+'_units']
            line = key+':'+tab+pv
            if units is None:
                pass
            else:
                line += ' '+units
            if isinstance(val,InfoArray):
                if (hasattr(val,'longName')):
                    line += " '"+val.longName+"'"
                else:
                    line += " '"+val.name+"'"
            else:
                pass
            rs.append(line)
        rs.append(bling)
        if aslist:
            return rs
        else:
            rs = '\n'.join(rs)
            return rs
       
    def plot(self,kind='image',fontsize=14,grid=False,oplot=False,scale='lin',fuzz=1e-9,**kwargs):
        """Display the InfoArray data as an image or lineout plot
        
        :param str kind: plot kind: 'image','lineout',or 'radial'
        :param int fontsize: for the titles, axis labels, and tick labels
        :param boolean grid: draw axis grid
        :param boolean oplot: whether or not to plot over a prevous graph (only for line plots)
        :param str scale: plot scale: 'lin','sqrt','log','loglog'
        :param float fuzz: a minimum value for log scale 
                    to keep the image contrast reasonable
                    (used only for 'image', scale='log')
        :param **kwargs: is sent on to dimg.show in the kind='image' case
        """
        scales = ['lin','sqrt','log','loglog']
        assert scale in scales,'scale must be one of '+scales
        n,m = self.shape
        if isinstance(self,np.matrix):
            valid_kinds = ['image']
            necessary_keys = set(['name'])
        else:
            valid_kinds = ['image','lineout','radial']
            necessary_keys = set(['units','dx','dx_units','name'])
        assert kind in valid_kinds,'kind must be one of '+valid_kinds
        keys = self.__dict__.keys()
        assert necessary_keys.issubset(set(keys)),'Object must have attributes '+str(list(necessary_keys))
        if hasattr(self,'longName'):
            name = self.longName
        else:
            name = self.name
        if isinstance(self,np.matrix):
            x0,y0 = 0,0
            if (hasattr(self,'axis_names')):
                xlabel = self.axis_names[1]
                ylabel = self.axis_names[0]
            else:
                xlabel = 'column index'
                ylabel = 'row index'
            origin = 'upper'
            xlim = [x0,x0+n]
            ylim = [y0+n,y0]
        else:
            dx = self.dx
            dx_units = self.dx_units
            if hasattr(self,'x0y0'):
                x0,y0 = self.x0y0
            else:
                x0,y0 = -(float(m)/2)*dx-dx/2, -(float(n)/2)*dx-dx/2
            origin = 'lower'
            xlim = [x0,x0+m*dx]
            ylim = [y0,y0+n*dx]
            if 'dy_units' in keys:
                dy_units = self.dy_units
            else:
                dy_units = self.dx_units
            if 'axis_names' in keys:
                xlabel = self.axis_names[0]+'  '+dx_units
                ylabel = self.axis_names[1]+'  '+dy_units
                if len(self.axis_names)>2:
                    rlabel = self.axis_names[2]+'  '+dx_units
                else:
                    rlabel = dx_units
            else:
                xlabel = dx_units
                ylabel = dy_units
                rlabel = dx_units
        
        data = np.array(self)
        if np.iscomplex(data).any():
            print '<Object.plot> data is COMPLEX - showing the absolute value'
            data = np.abs(data)
            name = name+' (ABSOLUTE VALUE)'
        
        if not oplot and kind is not 'image':
            plt.figure()
        
        if len(self.units) > 0:
            name += ', '+self.units
        
        if kind == 'radial':
            name += ' radial average'
            x = np.arange(xlim[0],xlim[1],dx)
            y = np.arange(ylim[0],ylim[1],dx)
            x,y = np.meshgrid(x,y)
            r = np.sqrt(x**2+y**2)
            bins = np.arange(0,(n/2)*dx,dx)
            i = np.digitize(r.flatten(),bins)
            c = np.bincount(i)
            q = np.zeros(n/2+1)
            for j,y in zip(i,data.flatten()):
                q[j] += y
            q /= c
            q = q[:-1]
            plt.plot(bins,q)
            if not oplot:
                ax = plt.gca()
                ax.set_xlabel(rlabel,fontsize=self.plotParams['fontsize.axis'])
                ax.set_ylabel(self.units,fontsize=self.plotParams['fontsize.axis'])
                ax.tick_params(axis='both',which='major',labelsize=self.plotParams['fontsize.tick'],direction='out',length=6,width=1,
                               labeltop=False,labelbottom=True,labelleft=True,labelright=False)
                plt.title(name,y=1.04)
            if scale == 'log':
                ax.set_yscale('log')
            if scale == 'loglog':
                ax.set_xscale('log')
                ax.set_yscale('log')
                
        if kind == 'lineout':
            y = data[n/2,:]
            x = np.arange(xlim[0],xlim[1],dx)
            plt.plot(x,y,**kwargs)
            if not oplot:
                ax = plt.gca()
                ax.set_xlabel(xlabel,fontsize=self.plotParams['fontsize.axis'])
                ax.set_ylabel(self.units,fontsize=self.plotParams['fontsize.axis'])
                ax.tick_params(axis='both',which='major',labelsize=self.plotParams['fontsize.tick'],direction='out',length=6,width=1,
                               labeltop=False,labelbottom=True,labelleft=True,labelright=False)
                plt.title(name,y=1.04)
            if scale == 'log':
                ax.set_yscale('log')

        if kind == 'image':
            if scale == 'log':
                data = np.log10(np.clip(np.abs(data),fuzz,np.inf))
            if scale == 'sqrt':
                data = np.sqrt(np.abs(data))
            dimg.show(data,extent = xlim+ylim,origin=origin,title=name,**kwargs)
            ax = plt.gca()
            if not oplot:
                ax.set_xlabel(xlabel,fontsize=self.plotParams['fontsize.axis'])
                ax.set_ylabel(ylabel,fontsize=self.plotParams['fontsize.axis'])
                ax.tick_params(axis='both',which='major',labelsize=self.plotParams['fontsize.tick'],direction='out',length=6,width=1)
                top,bottom,right,left = self.plotParams['axislabel.top'],self.plotParams['axislabel.bottom'],self.plotParams['axislabel.right'],self.plotParams['axislabel.left']
                ax.tick_params(axis='both',which='both',
                               top=top,bottom=bottom,right=right, left=left,
                               labeltop=top,labelbottom=bottom,labelleft=left,labelright=right)
                if hasattr(self,'sub'):
                    plt.subplots_adjust(top=0.84)
                    plt.suptitle(name, y=0.995, fontsize=self.plotParams['fontsize.title'])
                    plt.title(self.sub, y=1.08, fontsize=self.plotParams['fontsize.subtitle'])
                else:
                    plt.title(name,y=1.08, fontsize=self.plotParams['fontsize.title'])

        if not oplot:
            ax.tick_params(axis='both',which='minor',direction='out',length=4,width=1)
            plt.minorticks_on()
        
        plt.grid(grid)
        plt.draw()
        
    def graph(self,**kwargs):
        kwargs['kind'] = 'lineout'
        self.plot(**kwargs)
        
    def display(self,**kwargs):
        kwargs['kind'] = 'image'
        self.plot(**kwargs)
        
    def show(self,**kwargs):
        self.plot(**kwargs)
    
    def save(self,filename = None, group='/',file_type = None,clobber=False,verbose=False):
        """Write an InfoArray to a disk file. Only hdf5 and FITS file types
        are presently supported. The file type is derived
        from the filename extension (*.hdf5, *.fits); the
        keyword parameter file_type overrides.
        
        HDF5:
        
            Ceate a group within the hdf5 file
            (creating a new file if necessary) and store the
            data and metadata (InfoArray attributes) as datasets in the group,
            named 'data' and 'metadata' respectively.
            The group is named the same as the name of the InfoArray.
            Name uniqueness is checked within an existing group
            and will not overwrite an existing dataset of the same name.

            Storing into a subgroup is possible using the tree structure
            syntax of th group parameter.
        
        FITS:
        
            For FITS files, create the header from the
            InfoArray attributes and store the data as a single primary HDU.
            The filename is an optional parameter. The filename is based on
            the InfoArray's name (<name>.fits) if filename is not specivied.
        """
        assert self.name != ''
        valid_filetypes = {'hdf5':'hdf5','hd5':'hdf5',
                           'fits':'fits','fit':'fits'}
        if filename is None:
            filename = '%s.hdf5'%self.name
        basename,ext = os.path.splitext(filename)
        if ext == '':
            if file_type is not None:
                ext = file_type
            else:
                ext = 'hdf5'
        assert ext[1:] in valid_filetypes.keys()
        if verbose:
            print '<save> saving to file %s in %s format'%(filename,file_type)

        if file_type == 'hdf5':
            #metadata = encode(self.__dict__) # future encoder of dictionaries
            metadata = 'name: %s, dx: %r'%(self.name,self.dx)  # very primative for now
            name = '_'.join(self.name.split()) # replaces whitespace with _
            f = h5py.File(filename)
            if group in f:
                top = f[group]
            else:
                top = f.create_group(group)
            if name in top.keys():
                f.close()
                raise ValueError,'name "%s" not unique in %s'%(name,filename)
            grp = top.create_group(self.name)
            grp.create_dataset('data', data=self)
            grp.create_dataset('metadata',data=metadata)
            f.close()
        if file_type == 'fits':
            hdu = self._hdu()
            hdu.writeto(filename,clobber=clobber)
        else:
            raise ValueError, '%s is not a supported file_type'%file_type
    
    def _hdu(self):
        """Create a fits header data unit from the object
        data and metadata
        """
        hdu = pyfits.PrimaryHDU(self)
        hdr = hdu.header            
        for key in self.__dict__.keys():
            val = self.__dict__[key]
            if not isinstance(val,(str,int,long,float,bool)):
                continue
            if key+'_units' in self.__dict__:
                u = self.__dict__[key+'_units']
                hdr.append((key.upper()[:8],val,u),end=True)
            elif not key.endswith('_units'):
                hdr.append((key.upper()[:8],val),end=True)
        return hdu        
    
def load(filename,name=None, group='/',file_type = 'hdf5'):
    """Read in an InfoArray from a disk file.
    Type load('?') to get a list of supported file types
    """
    valid_filetypes = ['hdf5','zemax','fits']
    basename,ext = os.path.splitext(filename)
    if ext[1:] in valid_filetypes:
        file_type = ext[1:]
    if filename in ['help','?','file_types?']:
        return valid_filetypes
    if not os.path.isfile(filename):
        raise Exception,'file %s does not exist'%filename
    if name is None and file_type != 'fits':
        raise Exception,'<load> name must be provided'
    if file_type == 'hdf5':
        f = h5py.File(filename,'r')
        top = f[group]
        data = top[name]['data'].value
        meta = top[name]['metadata'].value
        f.close()
        # parse the metadata
        ml = [[y.strip() for y in x.split(':')] for x in meta.split(',')]
        d = {}
        for key,val in ml:
            try:
                val = float(val)
            except:
                val = val
            d[key.strip()] = val
        
        return InfoArray(data,**d)
    elif file_type == 'zemax':
        r = subprocess.call(['dos2unix',filename])
        f = open(filename,'r')
        lines = f.read()
        f.close()
        lines = lines.split('\n')
        
        # build up the header
        hdr = {}
        hdr['name'] = name
        hdr['explanation'] = lines[0]
        hdr['filename'] = lines[2].split(' : ')[1].strip()
        hdr['date'] = lines[4].split(' : ')[1].strip()
        # line 9 has dx
        line = lines[9].split()
        hdr['dx'] = float(line[3])
        hdr['dx_units'] = line[4][:-1].decode('utf8')
        
        # read the data. it starts at line 21 and ends at the second to last line
        rows = []
        for line in lines[21:-1]:
            row = map(float,line.split())
            rows.append(row)
        return InfoArray(rows,**hdr)
    elif file_type == 'fits':
        hdu = pyfits.open(filename)[0]
        data,hdr = hdu.data,hdu.header
        r = InfoArray(data,**hdr)
        keys = hdr.keys()
        for key in ['NAME','UNITS','DX','DX_UNITS']:
            kl = key.lower()
            if kl in r.__dict__:
                if not key.endswith('_UNITS'):
                    r.__dict__[kl] = hdr[key]
                else:
                    key = key[:-6]
                    r.__dict__[kl] = hdr.comments[key]
        if name is not None:
            r.name = name
        return r
    else:
        raise ValueError, '%s is not a supported file_type'%file_type

def whatsin(filename):
    """List the directory of an hdf file.
    """
    def printname(name):
        print name
    f = h5py.File(filename,'r')
    f.visit(printname)
    f.close()
    
test_oft = """
import info_array
reload(info_array)
InfoArray = info_array.InfoArray
import numpy as np
n = 256
a = np.zeros((n,n))
a = InfoArray(a,name='a',units='microns',dx=0.1,dx_units='meters',wavelength=0.5,wavelength_units='microns')
p = a.oft(upsamp=2)

psf = a.oft()
mtf = psf.mtf()
ena = psf.ena()
print 'ENA = %0.6f %s'%ena
"""

depth = lambda L: isinstance(L, list) and max(map(depth, L))+1

def show(arg,nmax=4):
    """
    Display a list or grid of InfoArrays all on one figure.
    
    argument:
        arg = and InfoArray, list of InfoArrays, or list of lists of InfoArrays (2d grid)
    
    keyword argument:
        nmax = the maximum number of plots on a figure in each direction
    """
    if isinstance(arg,InfoArray):
        arg.show()
        return
    assert isinstance(arg,list),'argument must be an InfoArray or a list'
    d = depth(arg)
    assert d <= 2
    figSize = np.array(plt.rcParams['figure.figsize'])
    if d==1:
        ncols = min(len(arg),nmax)
        figSize[0] = figSize[0]*ncols
    else:        
        nrows = min(len(arg),nmax)
        ncols = max([len(row) for row in arg])
        ncols = min(ncols,nmax)
        figSize[0] = figSize[0]*ncols
        figSize[1] = figSize[1]*nrows

    fig = plt.figure(figsize=figSize*.8)
    fignum = fig.number
    
    if d==1:
        for k in range(ncols):
            n = k+1
            sub = '1%d%d'%(ncols,n)
            mods = {'fontsize.title':12,
                    'axislabel.right':False,
                    'axislabel.top':False}
            arg[k].plotParams.update(mods)
            arg[k].show(fig=fignum,sub=sub)
        plt.subplots_adjust(top=0.85,bottom=0.1, left=0.10,right=0.95,hspace=0.4,wspace=0.2)
    else:
        for row in range(nrows):
            ncols_this_row = min(len(arg[row]),nmax)
            for col in range(ncols_this_row):
                n = row*ncols + col + 1
                sub = '%d%d%d'%(nrows,ncols,n)
                mods = {'fontsize.title':12,
                        'axislabel.right':False,
                        'axislabel.top':False}
                arg[row][col].plotParams.update(mods)
                arg[row][col].show(fig=fignum,sub=sub)
        plt.subplots_adjust(top=0.92,bottom=0.1, left=0.10,right=0.95,hspace=0.4,wspace=0.)

test_code = """
import numpy as np
from info_array import InfoArray,show
import matplotlib.pyplot as plt

names = [['a','b'],['c','d'],['e']]
al = []
for row in names:
    sl = []
    for name in row:
        sl.append(InfoArray(np.random.normal(size=(10,10)),name=name))
    al.append(sl)

show(al)
show(al[0])
"""

class InfoMatrix(np.matrix):
    """
    A sub class of `numpy.matrix <http://docs.scipy.org/doc/numpy-1.10.1/reference/generated/numpy.matrix.html>`_, with useful additional attributes like **name** and **units**.

    (See `numpy array subclassing <http://docs.scipy.org/doc/numpy-1.10.1/user/basics.subclassing.html>`_)
        
    to use::
    
        a = InfoMatrix(np.zeros((2,2)),name='blah')
        a.units = ...
    
    InfoArrays are containers in that they can contain any number of attributes. Useful ones are:
    
    :ivar str name: The name of the array
    :ivar str units: The units of the values in the array
    :ivar float dx: The physical size of one pixel in the array
    :ivar str dx_units: The units of the pixel size
    :ivar list axis_names: A list of strings containing the axis names
        The first and second are the horizontal and vertical axes correspondingly.
        An optionall third is the radial (for radial average plots).
    
    """
    def __new__(cls, input_array, name='',**kwargs):
        obj = np.asarray(input_array).view(cls)
        if type(input_array) is InfoArray:
            copyattrs(input_array,obj)
            setattr(obj,'name',name)
                
        else:
            obj.name = name
            obj.units = 'arbitrary'

        obj.__dict__.update(kwargs)
        return obj
    
    def __array_finalize__(self, obj):
        if obj is None: return
        if isinstance(obj,InfoMatrix):
            copyattrs(obj,self)
        
    def copy(self):
        """Make a deep copy of the InfoMatrix
        """
        obj = InfoMatrix(self.view(np.ndarray).copy())
        copyattrs(self,obj)
        return obj
    
    def pprint(self,full=False):
        """Pretty-print the InfoMatrix. This consists
        of reporting on the object's attributes.
        
        :param boolean full: if True, **pprint** also prints out the contents of the array.
            Default is to print only the "header" (attributes) information.

        """
        print self._pprint(full=full)
        
    def _pprint(self,print_width=40,aslist=False,rshort=False,full=False):
        cr = '\n'
        bling = 50*'='
        rs = []
        rs.append(bling)
        if hasattr(self,'longName'):
            rs.append(self.longName)
        else:
            if hasattr(self,'name') and self.name is not None:
                rs.append(self.name)
        rs.append(str(type(self)))
        rs.append(bling)
        rs.append('shape: '+str(self.shape))
        if rshort:
            if aslist:
                return rs
            else:
                rs = '\n'.join(rs)
                return rs
        if full:
            rs.append(str(self))
        tab = ' '*4
        d = self.__dict__
        keys = sorted(d.keys())
        for key in keys:
            if key.endswith('_units'): continue
            val = d[key]
            units = None
            if isinstance(val,(str,unicode)): pv = "'"+val+"'"
            else: pv = str(val)
            if len(pv) > print_width:
                pv = str(val.__class__)
                if isinstance(val,(np.ndarray,np.matrix)):
                    pv += ' '+str(val.shape)
            if key+'_units' in keys:
                units = d[key+'_units']
            line = key+':'+tab+pv
            if units is None:
                pass
            else:
                line += ' '+units
            if isinstance(val,(InfoArray,InfoMatrix)):
                if (hasattr(val,'longName')):
                    line += " '"+val.longName+"'"
                else:
                    line += " '"+val.name+"'"
            else:
                pass
            rs.append(line)
        rs.append(bling)
        if aslist:
            return rs
        else:
            rs = '\n'.join(rs)
            return rs
       
    def plot(self,kind='image',fontsize=14,grid=True,oplot=False,scale='lin',fuzz=1e-9,**kwargs):
        """Display the InfoMatrix data as an image or lineout plot
        
        :param str kind: plot kind: 'image','lineout',or 'radial'
        :param int fontsize: for the titles, axis labels, and tick labels
        :param boolean grid: draw axis grid
        :param boolean oplot: whether or not to plot over a prevous graph (only for line plots)
        :param str scale: plot scale: 'lin','sqrt','log','loglog'
        :param float fuzz: a minimum value for log scale 
                    to keep the image contrast reasonable
                    (used only for 'image', scale='log')
        :param **kwargs: is sent on to dimg.show in the kind='image' case
        """
        scales = ['lin','sqrt','log','loglog']
        assert scale in scales,'scale must be one of '+scales
        n,m = self.shape
        if isinstance(self,np.matrix):
            valid_kinds = ['image']
            necessary_keys = set(['name'])
        else:
            valid_kinds = ['image','lineout','radial']
            necessary_keys = set(['units','dx','dx_units','name'])
        assert kind in valid_kinds,'kind must be one of '+valid_kinds
        keys = self.__dict__.keys()
        assert necessary_keys.issubset(set(keys)),'Object must have attributes '+str(list(necessary_keys))
        if hasattr(self,'longName'):
            name = self.longName
        else:
            name = self.name
        if isinstance(self,np.matrix):
            x0,y0 = 0,0
            if (hasattr(self,'axis_names')):
                xlabel = self.axis_names[1]
                ylabel = self.axis_names[0]
            else:
                xlabel = 'column index'
                ylabel = 'row index'
            origin = 'upper'
            xlim = [x0,x0+n]
            ylim = [y0+n,y0]
        else:
            dx = self.dx
            dx_units = self.dx_units
            x0,y0 = -(float(m)/2)*dx-dx/2, -(float(n)/2)*dx-dx/2
            origin = 'lower'
            xlim = [x0,x0+m*dx]
            ylim = [y0,y0+n*dx]
            if 'dy_units' in keys:
                dy_units = self.dy_units
            else:
                dy_units = self.dx_units
            if 'axis_names' in keys:
                xlabel = self.axis_names[0]+', '+dx_units
                ylabel = self.axis_names[1]+', '+dy_units
                if len(self.axis_names)>2:
                    rlabel = self.axis_names[2]+', '+dx_units
                else:
                    rlabel = dx_units
            else:
                xlabel = dx_units
                ylabel = dy_units
                rlabel = dx_units
        
        data = np.array(self)
        if np.iscomplex(data).any():
            print '<Object.plot> data is COMPLEX - showing the absolute value'
            data = np.abs(data)
            name = name+' (ABSOLUTE VALUE)'
        
        if not oplot and kind is not 'image':
            plt.figure()
        
        if len(self.units) > 0:
            name += ', '+self.units
        
        if kind == 'radial':
            name += ' radial average'
            x = np.arange(xlim[0],xlim[1],dx)
            y = np.arange(ylim[0],ylim[1],dx)
            x,y = np.meshgrid(x,y)
            r = np.sqrt(x**2+y**2)
            bins = np.arange(0,(n/2)*dx,dx)
            i = np.digitize(r.flatten(),bins)
            c = np.bincount(i)
            q = np.zeros(n/2+1)
            for j,y in zip(i,data.flatten()):
                q[j] += y
            q /= c
            q = q[:-1]
            plt.plot(bins,q)
            if not oplot:
                ax = plt.gca()
                ax.set_xlabel(rlabel,fontsize=fontsize)
                ax.set_ylabel(self.units,fontsize=fontsize)
                ax.tick_params(axis='both',which='major',labelsize=fontsize,direction='out',length=6,width=1,
                               labeltop=False,labelbottom=True,labelleft=True,labelright=False)
                plt.title(name,y=1.04)
            if scale == 'log':
                ax.set_yscale('log')
            if scale == 'loglog':
                ax.set_xscale('log')
                ax.set_yscale('log')
                
        if kind == 'lineout':
            y = data[n/2,:]
            x = np.arange(xlim[0],xlim[1],dx)
            plt.plot(x,y)
            if not oplot:
                ax = plt.gca()
                ax.set_xlabel(xlabel,fontsize=fontsize)
                ax.set_ylabel(self.units,fontsize=fontsize)
                ax.tick_params(axis='both',which='major',labelsize=fontsize,direction='out',length=6,width=1,
                               labeltop=False,labelbottom=True,labelleft=True,labelright=False)
                plt.title(name,y=1.04)
            if scale == 'log':
                ax.set_yscale('log')

        if kind == 'image':
            if scale == 'log':
                data = np.log10(np.clip(np.abs(data),fuzz,np.inf))
            if scale == 'sqrt':
                data = np.sqrt(np.abs(data))
            dimg.show(data,extent = xlim+ylim,origin=origin,title=name,**kwargs)
            ax = plt.gca()
            if not oplot:
                ax.set_xlabel(xlabel,fontsize=fontsize)
                ax.set_ylabel(ylabel,fontsize=fontsize)
                ax.tick_params(axis='both',which='major',labelsize=fontsize,direction='out',length=6,width=1,
                               labeltop=True,labelbottom=True,labelleft=True,labelright=True)
                plt.title(name,y=1.08)

        if not oplot:
            ax.tick_params(axis='both',which='minor',direction='out',length=4,width=1)
            plt.minorticks_on()
        
        plt.grid(grid)
        plt.draw()
        
    def graph(self,**kwargs):
        kwargs['kind'] = 'lineout'
        self.plot(**kwargs)
        
    def display(self,**kwargs):
        kwargs['kind'] = 'image'
        self.plot(**kwargs)
        
    def show(self,**kwargs):
        self.plot(**kwargs)

def test():
    a = InfoArray([[1.,2],[3,4]],name='Test',dx=1,dx_units='arbitrary')
    a.units = 'integers'
    a.pprint()
    a.show()
    ac = a.copy()
    ac[0,0] = 2.
    assert ((ac - a) == InfoArray([[1.,0],[0,0]])).all()
    
    b = InfoMatrix([[1.,2],[3,4]],name='TestMatrix')
    b.units = 'integers'
    b.pprint()
    b.show()
    
    


# -*- coding: utf-8 -*-
"""  Structure function and average PSF calculator for astronomical imaging
  Reference:
  [1] Gavel, D. (2017). Point Spread Function for Ground Layer Adaptive Optics.
  arXiv:astro-ph.IM, 1706.00041. Retrieved from http://arxiv.org/abs/1706.00041
"""
import sys
import os
import platform
if 'VIRTUAL_ENV' in os.environ and platform.system() == 'Darwin':
    import matplotlib
    matplotlib.use('PDF') # to allow a virtual environment with no graphics back end
import matplotlib.pylab as plt
import numpy as np
from infoArray import InfoArray
import img
from StringIO import StringIO
from units import *
from oprint import pprint
import re
import warnings
import pandas as pd
import itertools
from collections import OrderedDict
import time
import h5py
import json
import random
from scipy.special import gamma, kv

# -------------------------- Classes --------------------------------

class Cn2_Profile(object):
    """The Profile class defines the atmospheric profile
    
    Instances contain
        Cn2 profile vs altitude
        the normalized Cn2 (normalized to sum to one)
        normalized altitude (normalized to mean height)
        computed r0, theta0
    Several valid profiles are pre-defined, but the caller can also
    specify with a 'name', 'Cn2' and 'h' arguments
    
    'HV' profiles requre keyword arguments 'r0' and 'theta0' and 'dh'
    Other profiles can optionally contain an 'r0' argument, in which case
    the profile is normalized (Cn2 is changed) to compute to that r0.
    """
    valid_profiles = ['CP Typical','MK Imaka','Cerro Pachon','Mauna Kea','Simple 1','Simple 2','Simple 2h_100']
    
    def __init__(self, profile='Simple 1', **kwargs):
        self.profile = profile
        self.make_profile(profile=self.profile,**kwargs)
        self.name = self.profile
        self.suppressWarnings = False
        if 'suppressWarnings' in kwargs:
            self.suppressWarnings = kwargs['suppressWarnings']
        self.warning = ''
    
    def pprint(self):
        pprint(self)
    
    def make_profile(self,profile,**kwargs):
        self.lam0 = self.lam = 0.55
        self.lam0_units = 'microns'
        self.lam_units = 'microns'
        self.Cn2_units = 'm^1/3 (Cn2*dh units)'
        self.r0_units = 'm'
        self.r00_units = 'm (at 0.55 microns)'
        self.h_units = 'm'
        self.sf_method = 'new'
        
        lam = self.lam0*units[self.lam0_units]

        if profile not in self.valid_profiles:
            self.profile_fromDatabase(profile)
            h,Cn2 = self.h,self.Cn2
        else:
            if profile in ['Cerro Pachon','Mauna Kea']:
                self.longName = profile
                rdir = os.path.dirname(__file__)
                filename = os.path.join(rdir,'data','cn2.csv')
                fp = open(filename)
                s = fp.read()
                fp.close()
                s = s.replace('\r','\n')
                u = np.genfromtxt(StringIO(s),delimiter=',',skip_header=3,dtype=float).transpose()
            if profile == 'CP Typical':
                self.longName = 'Gemini South site testing, typical profile'
                reference = """
                Andersen, D. R., Stoesz, J., Morris, S., Lloyd-Hart, M., Crampton, D., Butterley, T., ... Wilson, R. (2006).
                Performance Modeling of a Wide Field Ground Layer Adaptive Optics System.
                Publications of the Astronomical Society of the Pacific, 118(849), 1574-1590. http://doi.org/10.1086/509266
                """
                self.reference = re.sub(' +',' ',reference)
                h = np.array([1.,25,50,100,200,400,800,1600,5500])
                Cn2 = np.array([7.04, 2.25, 1.35, 1.24, 1.99, 2.87, 3.02, 1.75, 17.0])*10**(-14) # m^(1/3)
            if profile == 'MK Imaka':
                self.longName = 'Gemini North site testing, 50 percentile profile'
                reference = """
                Lai, O., Chun, M., Charles, J., Carlberg, R., Richer, H., Pazder, J., ... Vp, B. C. (2008).
                IMAKA : Imaging from Mauna KeA with an atmosphere corrected 1 square degree optical imager.
                Proc. of SPIE, 7015, 1-12. http://doi.org/10.1117/12.790114
                """
                self.reference = re.sub(' +',' ',reference)
                h = np.array(  [1., 70, 130., 200., 270, 330., 400., 450.])
                Cn2 = np.array([20.,2.,  0.6, 0.4, 0.35, 0.14, 0.12, 0.03 ])*10**(-14)
            if profile == 'Simple 1':
                self.longName = "Single layer atmosphere"
                if 'layer_height' in kwargs:
                    h1 = kwargs['layer_height']
                else:
                    h1 = 1.
                h =  [ h1 ]
                Cn2 = [ 1. ]
                r0 = 20*cm
                h = np.array(h)*km
                k = 2*np.pi/lam
                Cn2 = np.array(Cn2)*(1./0.423)*r0**(-5./3.)*k**(-2)
            elif profile == 'Simple 2':
                self.longName = 'Two layer atmosphere with 68% of Cn2 in ground layer'
                h = [0., 1.5] # km
                Cn2 = [0.68,0.32] # %
                r0 = 20*cm
                k = 2*np.pi/lam
                Cn2 = np.array(Cn2)*(1./0.423)*r0**(-5./3.)*k**(-2)
                h = np.array(h)*km
            elif profile == 'Simple 2h_100':
                self.longName = 'Two layer atmosphere with 68% of Cn2 at 100m above ground'
                h = [0.1, 1.5] # km
                Cn2 = [0.68,0.32] # %
                r0 = 20*cm
                k = 2*np.pi/lam
                h = np.array(h)*km
                Cn2 = np.array(Cn2)*(1./0.423)*r0**(-5./3.)*k**(-2)
            elif profile == 'Cerro Pachon':
                self.longName = 'Cerro Pachon, Chile, location of Gemini Observatory'
                h, Cn2 = u[0:2,0:7]
                r0 = 0.15
                k = 2*np.pi/lam
                Cn2 = Cn2/(0.423*k**2*r0**(5./3.))
                h = h*km
            elif profile == 'Mauna Kea':
                self.longName = 'Mauna Kea, Hawaii, location of Keck telescopes'
                h, Cn2 = u[3:5,0:10]
                r0 = 0.2
                k = 2*np.pi/lam
                Cn2 = Cn2/(0.423*k**2*r0**(5./3.))
                h = h*km
            
            self.h,self.Cn2 = h,Cn2
        
        self.r0_calc()
        self.r00 = r00 = self.r0
        if 'L0' in kwargs:
            self.L0 = kwargs['L0']
            self.L0_units = 'm'
            self.K0 = 2*np.pi/self.L0
            self.K0_units = 'm^-1'
        else:
            self.L0 = 'Infinity'
            self.K0 = 0.
        
        if 'lam' in kwargs:
            self.lam = kwargs['lam']
        
        if 'sf_method' in kwargs:
            self.sf_method = kwargs['sf_method']
            
        # If r0 is specified, it is r00, which is r0 at 0.55 microns.
        # Scale the profile, Cn2, so that it integrates to
        # the specified r00 instead of r0. Then, recalculate
        # r0 for the specified wavelength, lam.
        
        if 'r0' in kwargs:
            self.r00 = r00 = kwargs['r0']
        
        self.Cn2 = Cn2 = Cn2*(r00/self.r0)**(-5./3.)
        self.r0_calc()

        self.Cn2_bar = Cn2_bar = Cn2/np.sum(Cn2)
        h_bar = np.sum(Cn2_bar*h**(5./3.))**(3./5.)
        h_bar = [1,h_bar][h_bar>0]
        self.h_bar = h_bar
        self.h_bar_units = 'm'
        self.xi = h/h_bar
        if not hasattr(self,'r0'):
            self.r0_calc()
        if not hasattr(self,'theta0'):
            self.theta0_calc()
        
        # calculate the constants that will go
        # into calculating the structure function
        
        c0 = 8.*np.sqrt(2.)*((3./5.)*gamma(6./5.))**(5./6.)
        c1 = 16.*np.sqrt(2.)*((3./5.)*gamma(6./5.))**(5./6.)
        c2 = 5.*np.pi**(5./6.)
        c3 = 3.*gamma(11./6.)
        c4 = 5.*np.pi**(5./3.)*gamma(-5./6.)
        
        self.cons = [c0,c1,c2,c3,c4]
        self.make_PSF0()

    def profile_fromDatabase(self,profile,dbFile = 'Cn2_profiles.hd5'):
        """Read in a profile from a database of profiles
        argument:
            profile is a string of the form '<site> <tile>'. For example 'Armazones median' or 'Maunakea 13N 75%'
        keyword:
            dbFile is the HDF5 file that contains seeing profiles
        """
        rdir = os.path.dirname(__file__)
        dbFile = os.path.join(rdir,'data',dbFile)
        site,tile = profile.rsplit(' ',1)
        self.site,self.tile = site,tile
        self.databaseFile = dbFile
        store = pd.HDFStore(dbFile)
        # first, look for a specific table in the store
        name = 'df_%s'%site
        if name in store:
            df = store[name]
            metadata = store.get_storer(name).attrs.metadata
            store.close()
            self.databaseMetadata = metadata
            sel = df[(df.site == site) & (df.tile == tile)].iloc[0]
            h = np.array(sel.Altitude)
            Cn2 = np.array(sel.Cn2_dh)
        else:
            df = store['df']
            metadata = store.get_storer('df').attrs.metadata
            store.close()
            self.databaseMetadata = metadata
            sel = df[(df.Site == site) & (df.tile == tile)].iloc[0]
            h = list(sel.index[2:-2])
            h = map(lambda x: float(x.replace('GL','0m').replace('km','000').replace('m','')),h)
            h = np.array(h)
            Cn2 = np.array(list(sel.iloc[2:-2]))*1.e-14
        
        self.h,self.Cn2 = h,Cn2

    def r0_calc(self):
        """Compute r0 from the profile
        """
        Cn2,h,lam = self.Cn2,self.h,self.lam*microns
        k = 2*np.pi/lam
        self.r0 = (0.423*k**2*Cn2.sum())**(-3./5.)
        self.r0_units = 'm'
        return self.r0,self.r0_units
    
    def theta0_calc(self):
        """Compute theta0 from the profile
        """
        Cn2,h,lam = self.Cn2,self.h,self.lam*microns
        k = 2*np.pi/lam
        self.theta0 = (2.914*k**2*(Cn2*h**(5./3.)).sum())**(-3./5.)/arcsec
        self.theta0_units = 'arcsec'
        return self.theta0,self.theta0_units
    
    def _smat_prep_1_(self,c,dm_conjugate=0.,wfs_conjugate=0.,act_spacing=None):
        """first step in preparing to calculate the S matrix
        
        Argument:
            c - an instance of a Constellation object
        
        Keywords:
            dm_conjugate - conjugate altidude of the dm, meters
            wfs_conjugate - conjugate altitude of the wfs, meters
                       (results in a shift of thw wfs measurements)
            act_spacing - the actuator spacing, in meters (optional if modeling the DM Nyquist cutoff)
        """
        self.constellation = c
        assert isinstance(c,Constellation)
        theta0 = self.theta0*units[self.theta0_units]
        alpha = c.alpha
        
        self.dm_conjugate = dm_conjugate
        self.dm_conjugate_units = 'm'
        self.wfs_conjugate = wfs_conjugate
        self.wfs_conjugate_units = 'm'
        
        h_bar = self.h_bar
        zc, zs = self.dm_conjugate, self.wfs_conjugate
        r0,L0,K0 = self.r0,self.L0,self.K0
        xi_c = zc/h_bar
        xi_s = zs/h_bar
        eta = alpha / theta0
        if L0 == 'Infinity':
            l0 = L0
            k0 = 0.
        else:
            l0 = L0 / r0
            k0 = K0*r0
        
        self.xi_c, self.xi_s,c.eta = xi_c,xi_s,eta
        self.l0, self.k0 = l0,k0
        
        # normalized varaibles:
        #   Cn2_bar - Cn2 profile normalized to sum to one
        #   xi, xi_c, xi_s - altitude normalized to h_bar
        #   eta - guide star positions normalized to theta0
        #   l0 - outer scale normalized to r0
        #   k0 = K0 normalized to r0 
        #   mu - separation argument for structure function, normalized to r0
        
        if act_spacing is not None:
            mu_act = act_spacing/r0
            self.act_spacing = act_spacing
            self.act_spacing_units = self.r0_units
            self.mu_act = mu_act
            self.spatialFilter = True
        else:
            self.spatialFilter = False

    def _smat_prep_2_(self,r_lim='calc',n=512):
        """creates the S matrix
        """
        self.n_fine = n
        c = self.constellation
        xi_h,xi_c,xi_s = self.xi, self.xi_c, self.xi_s
        Cn2_bar = self.Cn2_bar
        r0 = self.r0
        eta,N = c.eta, c.N
        p314 = self.cons[0]**(-3./5.) # 0.314
        
        self.warning = ''
        if r_lim == 'calc':
            r_lim = self.r0*(n/8.)
            if r_lim < c.radius*arcmin*self.h_bar:
                dr = 2*r_lim/float(n)
                warn1 = 'WARNING: structure function extent %0.2f arcmin is less than constellation radius %0.2f arcmin at h_bar'%(r_lim,c.radius*arcmin*self.h_bar)
                warn2 = 'suggest increasing n (now %d) to >%d'%(n,2*c.radius*arcmin*self.h_bar/dr)
                self.warning = warn1+'\n'+warn2
                if not self.suppressWarnings:
                    print self.warning
        else:
            if r_lim/float(n) < self.r0/8.:
                warn1 = 'WARNING: structure function does not sample r0 well'
                warn2 = '  suggest increasing n (now %d) to >%d'%(n,8.*r_lim/self.r0)
                self.warning = warn1+'\n'+warn2
                if not self.suppressWarnings:
                    print self.warning
        self.r_lim = r_lim
        self.dr = dr = 2*r_lim/float(n)
        r = np.arange(-r_lim,r_lim,dr)
        mu = r/r0
        dmu = dr/r0
        mu_x,mu_y = np.meshgrid(mu,mu)
        S = np.zeros((N+1,N+1,n,n)) # Structure Function
        S0 = np.zeros((N+1,N+1,n,n)) # Correlation Function
        # [1] equation (59)
        for xi,Cn2b in zip(xi_h,Cn2_bar):
            for j in range(N+1):
                delta_j = [0,1][j==0] # Kroneker delta
                for jp in range(N+1):
                    delta_jp = [0,1][jp==0]
                    bx = p314*((eta[j,0] - eta[jp,0])*(xi-xi_s) + eta[0,0]*(xi_s - xi_c)*(delta_j - delta_jp))
                    by = p314*((eta[j,1] - eta[jp,1])*(xi-xi_s) + eta[0,1]*(xi_s - xi_c)*(delta_j - delta_jp))
                    ax = mu_x + bx
                    ay = mu_y + by
                    a = np.sqrt(ax**2 + ay**2)
                    b = np.sqrt(bx**2 + by**2)
                    if self.sf_method == 'original':
                        sfa = self._sf(a)
                    elif self.sf_method == 'new':
                        sfa = self._sf2([bx,by])
                    sfb = self._sf(b)
                    S[j,jp,:,:] += Cn2b*(sfa-sfb)
                    S0[j,jp,:,:] += -0.5*Cn2b*sfa
                        
        self.S_original = S.copy()
        if self.spatialFilter and self.mu_act != 0.:
            # [1] equation (67)
            # note, this is done only for the structure function
            # --- wrong ---
            # nyquist = (n/2)/(self.mu_act/dmu) # units: cycles per pixel
            # mu_r = np.sqrt(mu_x**2+mu_y**2) # units: r0
            # f = 1./(1. + np.exp((mu_r-nyquist)/(2*dmu)))
            # --- above is wrong, redoing! ---
            nyquist = 0.5/self.act_spacing # cycles per meter
            df = 1./(n*dr)
            fx = (np.arange(n)-n/2)*df
            fx,fy = np.meshgrid(fx,fx)
            fr = np.sqrt(fx**2+fy**2) +2.*df 
            f = 1./(1. + np.exp((fr - nyquist)/df))
            #f = np.where( mu_r < nyquist,1,0)
            #f =  np.where(np.logical_and( np.abs(mu_x) < nyquist, np.abs(mu_y) < nyquist ),1,0)
            h = 1 - f
            h = InfoArray(h,name='spatialFilter',
                          dx = df, dx_units = 'cycles per meter',
                          nyquist = nyquist, nyquist_units = 'cycles per meter',
                          act_spacing = self.act_spacing, spacing_units = self.r0_units)
            self.spatialFilter = h
            # ***Test*** h is a delta function, allowing all spatial frequencies
            # This should result in no AO correction
            #print 'Debug: Testing h = delta function'
            #h = 1.
            for j in range(N+1):
                for jp in range(N+1):
                    if j == 0 and jp == 0:
                        R = Q = 0.
                    else:
                        Q = img.ft(S[j,jp,:,:])*h
                        if j != 0 and jp != 0:
                            R = Q*h
                            R = img.ftinv(R).real
                            R = R[n/2,n/2] - R
                            Q = img.ftinv(Q).real
                            Q = Q[n/2,n/2] - Q
                            R = Q - R
                        else:
                            R = 0.
                            Q = img.ftinv(Q).real
                            Q = Q[n/2,n/2] - Q
                    S[j,jp,:,:] += Q + R
        self.S = S
        self.S0 = S0
        self.mu_x,self.mu_y = mu_x,mu_y
        self.dr = r[1]-r[0]
        self.dr_units = 'm'
        self.dmu = dmu
        self.r_x,self.r_y = np.meshgrid(r,r)
        self.r_x_units = self.r_y_units = 'm'
        
    def make_S(self,constellation,r_lim='calc',n=512,dm_conjugate=0.,wfs_conjugate=0.,act_spacing=None):
        """Make the S matrix given the guide star constellation
        
        Parameters:
            constellation - a constellation object
            r_lim - the upper limit to the shift, in meters.
                Default is to calculate this internally based on the
                guide star separation and telescope size
            n - the fine grid (number of fine pixels from -r_lim to +r_lim in the structure functios)
            dm_conjugate - conjugate altitude of the DM, in meters
            wfs_conjugate - conjugate altitude of the wavefront sensor, in meters
            act_spacing - the actuator spacing, in meters (optional if modeling the DM Nyquist cutoff).
        """
        self._smat_prep_1_(constellation,dm_conjugate,wfs_conjugate,act_spacing=act_spacing)
        self._smat_prep_2_(r_lim,n)
    
    def make_PSF(self,w, constellation=None, **kwargs):
        """Make the PSF (point spread function), and incidentally, the MTF (modulation transfer function)
        given a set of weights on the guide star measurements.
            
        Parameters:
            w - guide star weight vector (size equal to number of guidestars). The weights must sum to one.

        Keyword arguments:
            constellation - guide star constellation
            kwargs - keyword arguments to the Constellation constructor and to the S matrix generator (self.make_S)
        """
        # check to see if the constellation is provided. If so, create a new S matrix
        c_kwargs = ['ngs','radius','geometry','rot','field_point']
        s_kwargs = ['r_lim','n','dm_conjugate','wfs_conjugate','spatialFilter']
        
        if constellation is not None:
            fkwargs = dict_filter(kwargs,s_kwargs)
            self.make_S(constellation,**fkwargs)
        
        # if we don't have a constellation, then we need to make one using the keyword parameters
        if not hasattr(self,'constellation'):
            fkwargs = dict_filter(kwargs,c_kwargs)
            constellation = Constellation(**fkwargs)
            fkwargs = dict_filter(kwargs,s_kwargs)
            self.make_S(constellation,**fkwargs)
        
        # check to see if the field point needs adjusting. field_point is in arcmin
        if 'field_point' in kwargs:
            field_point = np.array(kwargs['field_point'])
            if not np.isclose(field_point,self.constellation.field_point).all():
                self.constellation.field_point = field_point
                self.constellation.alpha[0,:] = field_point*arcmin
            fkwargs = dict_filter(kwargs,s_kwargs)
            self.make_S(self.constellation,**fkwargs)

        N = self.constellation.N
        w = np.array(w)
        if len(w) == N:
            wsum = w.sum()
            w = np.append([-1],w)
        elif len(w) == N+1:
            wsum = w[1:].sum()
        
        if len(w) != N+1:
            raise Exception,'length of w must equal the number of guidestars in the constellation %d'%N
        if not np.isclose(wsum,1.):
            raise Exception,'sum of weights =%0.2f; it is supposed to equal one'%wsum
        
        self.w = w
        S,S0 = self.S,self.S0
        N1,N1,n,n = S.shape
        lam,dr = self.lam*microns, self.dr
        
        D = np.zeros((n,n)) # Structure Function
        C = np.zeros((n,n)) # Correlation Function
        for j in range(N1):
            for jp in range(N1):
                D += w[j]*w[jp]*S[j,jp]
                C += w[j]*w[jp]*S0[j,jp]
        
        Dtele = 20.*self.r0
        u = np.sqrt(self.r_x**2 + self.r_y**2)/Dtele
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            tau = (2./np.pi)*(np.arccos(u) - u*np.sqrt(1-u**2))
        tau = np.nan_to_num(tau)
        MTF = np.exp(-0.5*D)*tau
        fpx,fpy = np.round(self.constellation.field_point,2)
        name = r' $C_n^2$="%s"  $N_{gs}=%d_{\rm %s}$, $\Theta=%g^\prime$, $\theta=(%g,\,%g)$, $\lambda=%g\mu m$'%(self.profile,N,self.constellation.geometry,self.constellation.radius,fpx,fpy,np.round(self.lam,2))
        MTF = InfoArray(MTF,'MTF',sub=name,dx=self.dr,dx_units='meters',axis_names=[r'$r_x$',r'$r_y$'],
                        profile=self.profile,ngs=self.constellation.N)
        PSF = oft(MTF)
        dth = lam/(n*dr)/arcsec
        PSF = InfoArray(PSF,'PSF',sub=name,dx=dth,dx_units='arcsec',
                        wavelength=lam/microns,wavelength_units='microns',
                        r0 = self.r0, r0_units = self.r0_units,
                        axis_names=[r'$\theta_x$',r'$\theta_y$'],
                        profile = self.profile,
                        ngs=self.constellation.N)
        th = np.arange(-n/2,n/2)*dth
        # Normalize the PSF to ~integrate~ to one
        PSF /= PSF.sum()*dth**2
        # calculate King's Equivalent Noise Area
        # (King, I. R. (1983). Accuracy of Measurement of Star Images on a Pixel Array. Publications of the Astronomical Society of the Pacific, 95(February), 163–168.)
        PSF.ena = dth**2/np.sum((PSF/PSF.sum())**2)
        PSF.ena_units = 'arcsec^2'
        PSF.ena_r = np.sqrt(PSF.ena/np.pi)
        PSF.ena_r_units = 'ENA radius, arcsec'
        
        self.dth, self.th = dth, th
        self.dth_units = self.th_units = 'arcsec'
        #PSF[n/2,n/2] = PSF[n/2,n/2+1] + (PSF[n/2,n/2+1]-PSF[n/2,n/2+2])/3.
        
        self.D,self.C = D,C
        self.MTF,self.PSF,self.tau = MTF,PSF,tau
    
    def _sf(self,r):
        """The finite outer scale structure function
        """
        c0,c1,c2,c3,c4 = self.cons
        k0 = self.k0
        if k0 == 0:
            return c0*r**(5./3.)
        eps = 1.e-16
        r += eps
        sf = c1*(c2*(k0*r)**(5./6.)*kv(5./6.,2*k0*np.pi*r)-c3) / (c4*k0**(5./3.))
        return sf
    
    def _sf2(self,b):
        """Generate finite outer scale structure function
        at a shift D(r+b), b = [bx,by], using a
        pre-computed structure function D(r)
        """
        n = self.n_fine
        if not hasattr(self,'D2'):
            r = np.arange(-self.r_lim,self.r_lim,self.dr)
            mu = r/self.r0
            mu_x,mu_y = np.meshgrid(mu,mu)
            mu = np.sqrt(mu_x**2+mu_y**2)
            self.D2 = self._sf(mu)
        D2,dmu = self.D2,self.dmu
        b = np.array(b)
        s = np.round(b/dmu).astype(int)
        indx = range(s[0],s[0]+n)
        indy = range(s[1],s[1]+n)
        sf = D2.take(indx,axis=0,mode='clip')
        sf = sf.take(indy,axis=1,mode='clip')
        return sf
    
    def make_PSF0(self,n=512,r_lim='calc'):
        """Make the diffration-limited and seeing limited PSFs and MTFs
        """
        if hasattr(self,'n_fine'):
            n = self.n_fine
        else:
            self.n_fine = n
            r0,L0,K0 = self.r0,self.L0,self.K0
            if r_lim == 'calc':
                r_lim = r0*(n/8.)
            else:
                if r_lim/float(n) < r0/8.:
                    print 'WARNING: structure function does not sample r0 well'
                    print '  suggest increasing n (now %d) to >%d'%(n,8.*r_lim/r0)
            dr = 2*r_lim/float(n)
            r = np.arange(-r_lim,r_lim,dr)
            mu = r/r0
            mu_x,mu_y = np.meshgrid(mu,mu)
            if L0 == 'Infinity':
                l0 = L0
                k0 = 0
            else:
                l0 = L0/r0
                k0 = K0*r0
            self.mu_x,self.mu_y = mu_x,mu_y
            self.dr = r[1]-r[0]
            self.dr_units = 'm'
            self.dmu = self.dr/r0
            self.r_x,self.r_y = np.meshgrid(r,r)
            self.r_x_units = self.r_y_units = 'm'
            self.l0,self.k0 = l0,k0

        xi, Cn2_bar = self.xi, self.Cn2_bar
        mu_x, mu_y = self.mu_x, self.mu_y
        lam = self.lam*microns
        n,dr = self.n_fine,self.dr
        
        D = np.zeros((n,n))
        mu = np.sqrt(mu_x**2+mu_y**2)

        sfmu = self._sf(mu)
        for xi,Cn2b in zip(xi,Cn2_bar):
            D += Cn2b*sfmu
        self.D_seeing = D
        MTF = np.exp(-0.5*D)
        name = r' $C_n^2=$"%s"  $r_0=%gm$, $\lambda=%g\mu m$'%(self.profile,np.round(self.r0,2),np.round(self.lam,2))
        MTF = InfoArray(MTF,r'${\rm MTF}_{\rm seeing}$',sub=name,dx=self.dr,dx_units='meters',
                        axis_names=[r'$r_x$',r'$r_y$'])
        PSF = oft(MTF)
        dth = lam/(n*dr)/arcsec
        PSF = InfoArray(PSF,r'${\rm PSF}_{\rm seeing}$',sub=name,dx=dth,dx_units='arcsec',
                        wavelength=lam/microns,wavelength_units='microns',
                        r0 = self.r0, r0_units = self.r0_units,
                        axis_names=[r'$\theta_x$',r'$\theta_y$'])
        # Normalize the PSF to ~integrate~ to one
        PSF /= PSF.sum()*dth**2
        # calculate King's Equivalent Noise Area
        # (King, I. R. (1983). Accuracy of Measurement of Star Images on a Pixel Array. Publications of the Astronomical Society of the Pacific, 95(February), 163–168.)
        PSF.ena = dth**2/np.sum((PSF/PSF.sum())**2)
        PSF.ena_units = 'arcsec^2'
        PSF.ena_r = np.sqrt(PSF.ena/np.pi)
        PSF.ena_r_units = 'ENA radius, arcsec'

        self.MTF_seeing, self.PSF_seeing = MTF,PSF

        Dtele = 20.*self.r0
        u = np.sqrt(self.r_x**2 + self.r_y**2)/Dtele
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            tau = (2./np.pi)*(np.arccos(u) - u*np.sqrt(1-u**2))
        tau = np.nan_to_num(tau)
        tau = InfoArray(tau,r'${\rm MTF}_{\rm dl}$',sub=name,dx=self.dr,dx_units='meters',
                        axis_names=[r'$r_x$',r'$r_y$'])
        self.MTF_dl = tau
        PSF_dl = oft(tau)
        self.PSF_dl = InfoArray(PSF_dl,r'${\rm PSF}_{\rm dl}$',sub=name,dx=dth,dx_units='arcsec',
                        axis_names=[r'$\theta_x$',r'$\theta_y$'])
    
    def profile_plot(self,altscale='log',cn2scale='log',drawstyle='steps'):
        """plot the profile and cumulative Cn2 profile
        Keywords:
            altscale - altitude scale; can be 'linear' or 'log'
            cn2scale - Cn2 axis scale; can be 'linear' or 'log'
            drawstyle - draw style for the cumulative graph: can be 'default' or 'steps'
            (Cn2 draw style is bar, with dot at end)
        """
        h,Cn2,Cn2_bar = self.h, self.Cn2, self.Cn2_bar
        cumProfile = []
        a = 0.
        for c in Cn2_bar:
            a += c
            cumProfile.append(a)
        cumProfile = np.array(cumProfile)
    
        plt.subplot(121)
        if altscale == 'log':
            plt.barh(h,Cn2/10**(-14),log=True,height=h*.05)
        else:
            plt.barh(h,Cn2/10**(-14),height = h.max()*.005)
        plt.plot(Cn2/10**(-14),h,'bo')
        plt.xlabel(r'$C_n^2$, $\times 10^{-14} m^{1/3}$')
        plt.ylabel('Altitude, meters')
        plt.yscale(altscale)
        plt.xscale(cn2scale)
        plt.ylim(0.9,10000.)
        plt.grid('on')
    
        plt.subplot(122)
        cumProfile = np.append([0],cumProfile)
        h = np.append([.9],h)
        plt.plot(cumProfile*100,h,linewidth=3,drawstyle=drawstyle)
        plt.xlabel(r'Accumulated percentage of $C_n^2$')
        plt.ylabel('Altitude, meters')
        plt.ylim(0.9,10000.)
        plt.yscale(altscale)
        plt.xlim(0,100)
        plt.grid('on')
        plt.subplots_adjust(top=0.92,bottom=0.1, left=0.15,right=0.95,hspace=0.25,wspace=0.35)        
    
        plt.suptitle(r'$C_n^2$ profile: "%s"'%self.profile,fontsize=14)

    def graph(self, which='structure',type='line',over=False,linestyle=None,color=None,linewidth=None,label='',dots=False, normalized_r=False):
        """graph one of the characteristic functions, Structure function or Correlation function.
        
        keyword:
            which - 'structure' or 'correlation'
            type - the type of graph, either 'line' (lineout) or 'grey' (2-d image greyscale)
            over - for lineout, this allows multiple overplots. First call should be False, then True for subsequent overplots
            label - if multiple overplots, the text for a legend (call plt.legend() after the last overplot)
        """
        assert which in ['structure','correlation']
        assert type in ['line','grey']
                    
        n,n = self.D.shape
        title = 'Profile: "%s", $N_{GS} =$ %d %s'%(self.name,self.constellation.N,self.constellation.geometry)
        if which == 'structure':
            P = self.D
            ylabel = r'Structure Function ${\cal D}_{\psi_r}(r)$, radians$^2$ @ $\lambda = %0.2f\, {\mu}m$'%(self.lam*units[self.lam_units]/microns)
        elif which == 'correlation':
            P = self.C
            ylabel = r'Correlation Function ${\cal C}_{\psi_r}(r)$, radians$^2$'
        if normalized_r:
            xlabel = r'normalized separation, $\mu=r/r_0$'
        else:
            xlabel = r'$r$, meters'
        if type == 'line':
            P = P[n/2,n/2+1:]
            r = self.r_x[n/2,n/2+1:]
            mu = self.mu_x[n/2,n/2+1:]
            if normalized_r:
                r = mu
            if which == 'structure' and not over:
                label0 = r'$6.88(r/r_0)^{5/3}$'
                plt.plot(r,6.88*mu**(5./3.),'k--',label=label0)                
            line, = plt.plot(r,P,linestyle=linestyle,color=color,linewidth=linewidth,label=label)
            color = line.get_color()
            if dots: plt.plot(r,P,'.',color=color)
            if which == 'structure':
                plt.xscale('log')
                plt.yscale('log')
            if not over:
                plt.grid('on',which='both')
                plt.title(title)
                plt.xlabel(xlabel)
                plt.ylabel(ylabel)
        elif type == 'grey':
            print 'type=grey not implemented yet'
    
    def showPSF(self,scale='log',which='both'):
        """Display PSF, and optionally, PSF_seeing, with (optionally) the Equiavlent Noise Area circled
        
        Keyword:
            scale = 'linear' or 'log'
            which = 'AO', 'seeing', or 'both'
        """
        if which == 'AO': PSFset = [self.PSF]
        if which == 'seeing': PSFset = [self.PSF_seeing]
        if which == 'both': PSFset = [self.PSF,self.PSF_seeing]
        
        for PSF in PSFset:
            if scale == 'log':
                PSF = np.log(PSF)
            PSF.show()
            n,m = PSF.shape
            dx = PSF.dx
            r = PSF.ena_r
            th = np.linspace(-np.pi,np.pi)
            x,y = r*np.sin(th),r*np.cos(th)
            plt.plot(x,y,'r')
            plt.xlim(-dx*n/2,dx*n/2)
            plt.ylim(-dx*m/2,dx*m/2)
        
class Constellation(object):
    """Defines the locations of the guide stars
    
    arguments:
        ngs = number of guide stars
        radius = radius of guide star constellation, arcmin
        geometry = 'circle', 'wheel', 'grid'
            ('wheel' is a circle with one guidestar in the middle)
            default is 'wheel'
        rot = rotation of the constellation on-sky, degrees
        field_point = position of the science field position, arcmin
    """
    def __init__(self,ngs=1,radius=10,geometry='wheel',rot=0., field_point=[0.,0.]):
        N = ngs
        radius = radius*arcmin
        rot = rot*degrees
        self.name = 'Guide star constellation'
        alpha = np.zeros((N+1,2))
        if geometry == 'circle':
            for k in range(1,N+1):
                theta = k*2*np.pi/float(N) + rot
                alpha[k,:] = [radius*np.cos(theta),radius*np.sin(theta)]
        elif geometry == 'wheel':
            alpha[1,:] = [0,0]
            for k in range(2,N+1):
                theta = k*2*np.pi/float(N-1) + rot
                alpha[k,:] = [radius*np.cos(theta),radius*np.sin(theta)]
        elif geometry == 'grid':
            assert N in [4,9,16] # only allowable grid patterns
            
            n = np.sqrt(N).astype(int)
            q = np.arange(n) - (n-1)/2.
            spacing = (2.*radius/np.sqrt(2))/float(n-1)
            k = 1
            for x in q:
                for y in q:
                    r = np.sqrt(x**2+y**2)*spacing
                    theta = np.arctan2(x,y) + rot
                    alpha[k,:] = [r*np.cos(theta),r*np.sin(theta)]
                    k = k+1
        self.field_point = np.array(field_point)
        alpha[0,:] = np.array(field_point)*arcmin
        
        self.N = N
        self.N_units = 'number of guide stars'
        self.geometry = geometry
        self.radius = radius/arcmin
        self.radius_units = 'arcmin'
        self.rotation = rot/degrees
        self.rotation_units = 'degrees'
        self.alpha = alpha
        self.alpha_units = 'radians'
        self.field_point = field_point
        self.field_point_units = 'arcmin'
    
    def graph(self,field_points = None,colors={}):
        """Graph the constellation of guidestars and the field point
        Keyword:
            Field_points = a list of [x,y] field points to graph instead of the one associated with the constellation
        """
        default_colors = {'guidestar':'orange',
                  'field point':'cyan',
                  'background':'black',
                  'axes':'grey',
                  'grid':'darkred'}
        default_colors.update(colors)
        colors = default_colors
        x,y = self.alpha.transpose()/arcmin
        r = max(np.abs(x).max(),np.abs(y).max())*1.1
        plt.plot(x[1:],y[1:],'*',color=colors['guidestar'],label='guidestar')
        if field_points is None:
            plt.plot(x[0],y[0],'o',color=colors['field point'],label='field point')
        else:
            label = 'field point'
            for point in field_points:
                plt.plot(point[0],point[1],'x',color=colors['field point'],label=label)
                label = None
        plt.xlim(-r,r)
        plt.ylim(-r,r)
        plt.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        plt.axes().set_aspect('equal')
        plt.grid('on',color=colors['grid'])
        plt.plot([-r,r],[0,0],colors['axes'])
        plt.plot([0,0],[-r,r],colors['axes'])
        #plt.gca().set_facecolor(colors['background'])
        plt.xlabel(r'$\theta_x$, arcmin')
        plt.ylabel(r'$\theta_y$, arcmin')
        plt.title(self.name)

    def pprint(self):
        pprint(self)

# --------------------------- Helper Routines -------------------------------

def progress_bar(n,N):
    """This creates and maintains a progress bar to show the status of long computations
    Arguments:
        n = the number of operations performed so far
        N = the total number of operations to be performed
    """
    pc = int(np.round(100*float(n)/float(N)))
    N_max = 50
    d = N//N_max + 1
    N = N//d
    n = n//d
    sys.stdout.write('\r')
    fmt = '[%%-%ds]' % N
    ostr = (fmt+' %d%%') % ('='*n,pc)
    sys.stdout.write(ostr)
    sys.stdout.flush()
    
def available(show=True):
    """List the available profiles, including those in databases
    """
    rdir = os.path.dirname(__file__)
    dbFile = os.path.join(rdir,'data','Cn2_profiles.hd5')
    store = pd.HDFStore(dbFile)
    qlist = zip(list(store['df']['Site']),list(store['df']['tile']))
    qlist += [('MK2009','25%'),('MK2009','50%'),('MK2009','75%')]
    store.close()
    qlist = map(lambda x: '%s %s'%(x[0],x[1]),qlist)
    if show:
        print '---- In-code profiles: ----'
        for p in Cn2_Profile.valid_profiles:
            print p
        print '\n---- Profiles in %s ----'%dbFile
        for q in qlist:
            print q
    else:
        return Cn2_Profile.valid_profiles+qlist

def run_cases(testmode = False, HDF=None, **kwargs):
    """This is a generic multi-loop over PSF calculations with different parameters.
    Returns a pandas DataFrame of results, along with metadata dictionary (as df.metadata).
    
    Keyword:
        HDF - the name of the HDF file to store the PSFs into. A unique ID is created and
              entered into the database in an additional column. Default: None
              
    Typical parameters, passed via kwargs, include:
        theta_set - list of constellation opening angles (arcmin)
        lam_set - list of wavelengths (microns)
        site_set - list of site names - these are used to look up Cn2 profiles in a database
    
    Other optional parameters are:
        alpha - marching out angle for field points on a line (degrees)
        radius_minimum - constrains the minimum size of a guide star constellation (arcmin)
        N - number of guidestars in the constellation
        geom - geometry of guidestar constellation ('whhel', or 'circle', with 'wheel' meaning one GS at the center)
        n - size of fine grid for the structure function calculation
    """
    # parameters to vary
    default_iters = OrderedDict(
        [
            ('theta_set', [0., 0.5, 1., 2.5, 3.5, 5.0]),
            ('lam_set',[0.55]),
            ('site_set',['MK Imaka','Armazones 25%','Armazones median','Armazones 75%',
                         'Maunakea 13N 25%','Maunakea 13N median','Maunakea 13N 75%',
                         'ORM 25%','ORM median','ORM 75%']),
        ]
    )
    # columns of the returned database:
    cols = ['Profile','telescope','DM_conjugate','wfs_conjugate','wavelength',
            'r00','r0','theta0','h_bar','constellation_radius',
            'field_angle_r','field_angle_theta','ENA_R_seeing','ENA_R','warning']
    
    if HDF is not None:
        cols = cols + ['file','date','time','uid','key']

    default_variables = [
        ('alpha',60.),
        ('radius_minimum',1.0),
        ('N',3),
        ('n',256),
        ('geom','circle'),
        ('dm_conjugate',None),
        ('wfs_conjugate',None),
        ('conjugate_dict',None),
        ('telescope',None),
        ]  # if this list is modified, change line ===>(A) below
    default_variables = OrderedDict(default_variables)
    # ------- sort out the arguments ---
    for d in [default_iters,default_variables]:
        keys = d.keys()
        dd = dict([(x,kwargs[x]) for x in keys if x in kwargs])
        d.update( dd )
    
    caseload = default_iters.keys()
    caseload = map(lambda x: default_iters[x],caseload)
    the_cases = list(itertools.product(*caseload))
    
    vns = default_variables.keys()
    
    # (A)===> change this line if default_variables is changed:
    (alpha,radius_minimum,N,n,geom,dm_conjugate,wfs_conjugate,conjugate_dict,telescope) = map(lambda x: default_variables[x],vns)
    
    telescopes = {'MK':'TMT',
                  'Maunakea':'TMT',
                  'ORM':'TMT',
                  'Armazones':'E-ELT'}
    telescope_dict = {}
    for site in default_iters['site_set']:
        key = site.split()[0]
        telescope_dict[site] = telescopes[key] if key in telescopes else ''
    
    if conjugate_dict is None:
        conjugate_dict = {'TMT':[-390]*2,
                          'E-ELT':[556]*2,
                          'Keck':[-126]*2,
                          '':[0]*2 }
    if dm_conjugate is not None:
        for key in conjugate_dict:
            conjugate_dict[key][0] = dm_conjugate
    if wfs_conjugate is not None:
        for key in conjugate_dict:
            conjugate_dict[key][1] = wfs_conjugate

    vr = np.array([np.cos(alpha*degrees),np.sin(alpha*degrees)])
    
    # open the database file
    if HDF is not None:
        dfile = h5py.File(HDF) # append or create new

    k,count = 0,len(the_cases)
    rows = []
    progress_bar(k,count)
    first = True
    
    for theta,lam,site in the_cases:
        telescope = default_variables['telescope']
        if telescope is None:
            telescope = telescope_dict[site]
        dm_conjugate, wfs_conjugate = conjugate_dict[telescope]
        radius = max(abs(theta)*1.2,radius_minimum)
        if not testmode or first:
            
            # ======== code goes here ========
            
            a = Cn2_Profile(site,lam=lam,suppressWarnings=True)
            r_lim = a.r0*n/4
            w = np.zeros(N+1)
            w[0] = -1
            w[1:] = np.ones(N)/float(N)
            radius = max(abs(theta)*1.2,radius_minimum)
            field_point = theta*vr
            constellation = Constellation(ngs=N,geometry=geom,radius=radius)
            a.make_PSF( w, constellation, n=n,
                       field_point = field_point,
                       dm_conjugate = dm_conjugate,
                       wfs_conjugate = wfs_conjugate)
            
            # ================================
            
            row = [a.name, telescope, dm_conjugate, wfs_conjugate, a.lam,
                   a.r00, a.r0,  a.theta0,  a.h_bar,   radius,
                   theta, alpha, a.PSF_seeing.ena_r,  a.PSF.ena_r, a.warning] # like to add FWHM and ellipticity
            first = False
        else:

            row = [site, telescope, dm_conjugate, wfs_conjugate, '',
                   '', '',  '',  '',   radius,
                   theta, alpha, '',  '', '']

        uid,dat,tim = uniq_id(5)
        key = '%s_%sT%s'%(uid,dat,tim)
        if HDF is not None:
            grp = dfile.create_group(key)
            grp.create_dataset('data',data=a.PSF)
            attrs = a.PSF.__dict__
            grp.create_dataset('attrs',data=json.dumps(attrs))
            row = row + [HDF,dat,tim,uid,key]
            hdr = dict(pd.Series(row,cols))
            grp.create_dataset('header',data = json.dumps(hdr))
        rows.append(row)
        k += 1
        progress_bar(k,count)
    
    progress_bar(count,count)
    print '\nDone'
    df = pd.DataFrame(rows,columns = cols)
    df.metadata = {
        'code':'paper2.run_cases',
        'iters':default_iters, 'variables':default_variables,
        'time':time.strftime('%Y/%m/%dT%H:%M:%S')
    }
    
    if HDF is not None:
        dfile.close()
    
    return df

def uniq_id(n):
    """Create a unique id string ('key') containing n random characters
    followed by a timestamp.
    
    Returns the random character string, the datestamp, and the timestamp
    as a tuple of strings
    """
    chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890"
    uid = ''.join(random.choice(chars) for _ in range(n))
    dat,tim = str(pd.Timestamp('now')).split(' ')
    return uid,dat,tim
    
def dfSave(df,name,hfile = 'paper3.h5'):
    """Save a pandas database in an HDF file.
    This will also save the metadata.
    It will not overwrite existing datasets in the HDF file, unless told to.
    
    Argument:
        df - the pandas dataset
        name - the name to give it in the HDF Store file
    """
    store = pd.HDFStore(hfile)
    if name not in store:
        store[name] = df
        if hasattr(df,'metadata'):
            store.get_storer(name).attrs.metadata = metadata
    else:
        print 'WARNING - %s is in %s. opted not to write over'%(name,hfile)
    store.close()
    
def dfRead(name,hfile='paper3.h5'):
    """Read a pandas database from an HDF file.
    """
    store = pd.HDFStore(hfile)
    df = store[name]
    q = store.get_storer(name)
    if hasattr(q,'metadata'):
        df.metadata = q.metadata
    store.close
    return df

def oft(A):
    """Optical Fourier Transform
    """
    return np.abs( np.fft.ifftshift( np.fft.fft2( np.fft.fftshift(A))))

def dict_filter(dict,key_list):
    """Return a dictionary derived from dict but having only keys from key_list
    """
    r = {}
    for key in dict:
        if key in key_list:
            r[key] = dict[key]
    return r

def fwhm(f):
    """determine full-width-half-max of a lineout
    """
    k = np.argmax(f)
    x = f.max()
    u = np.abs(f[0:k]-0.5*x)
    k1 = np.argmin(u)
    u = np.abs(f[k:]-0.5*x)
    k2 = np.argmin(u)+k
    return k2-k1

def fwhm2(f,method='area'):
    """find the full-width-half-maxima of a 2d blob
    """
    n,m = f.shape
    fmax = f.max()
    b = (f > fmax/2).astype(int)
    if method == 'area':
        cb = b.sum()
        r = np.sqrt(cb/np.pi)*2.
        r_major,r_minor = r,r
    if method == 'slice':
        return (fwhm(f[n/2,:]),fwhm(f[:,m/2]))
    else:
        iy,ix = np.where(f > fmax/2.)
        cx = np.average(ix.astype(float))
        cy = np.average(iy.astype(float))
        x = ix - cx
        y = iy - cy
        r = np.sqrt(x**2+y**2)
        k = np.argmax(r)
        theta = np.arctan2(x[k],y[k])
        rp = np.arange(0,r.max())
        xp,yp = rp*np.cos(theta+np.pi/2),rp*np.sin(theta+np.pi/2)
        ixp,iyp = (xp+cx).astype(int),(yp+cy).astype(int)
        c = b.flat[ixp+iyp*n]
        xp,yp = xp[np.where(c)],yp[np.where(c)]
        rp = np.sqrt(xp**2+yp**2)
        r_major = r.max()
        r_minor = rp.max()
    return (2*r_major,2*r_minor)

# --------------------------- Tests -------------------------------------

def test():
    df = run_cases(testmode=True)
    print df



#!/usr/bin/env python
"""Dimg is the Display IMaGe package
wrapping some most-used parts of the matplolib library
and pyfits library, and making graphics more interactive.

Author: Don Gavel
"""
import os
import time

try:
    machine = os.environ['HOSTNAME']
except:
    machine = 'not a unix host (probably a mac)'

if (machine != 'rtc.ucolick.org'):
    import matplotlib
    # if (matplotlib.get_backend() != 'TkAgg'):
    #     matplotlib.use('TkAgg',warn = False)
    from mpl_toolkits.mplot3d import Axes3D
    from matplotlib import cm
    from matplotlib.ticker import LinearLocator, FormatStrFormatter
    import matplotlib.pyplot as plt
    plt.ion()

import numpy as np
import readfits as fits
import matplotlib.animation as animation

def surface(z,**kwargs):
    dimg(z,surface=True,**kwargs)
    
def contour(z,**kwargs):
    if 'image' not in kwargs.keys():
        kwargs['image'] = False
    dimg(z,contour=True,**kwargs)
    
def tv(z,grid=True,cmap='bone',origin='lower',dx=1):
    """New simplified version of dimg
    Displays only 2-D images
    X and Y coordinates start at 0,0 lower-left of 0,0 pixel and end at n,n upper right of n-1,n-1 pixel
    Display origin is always lower left
    Always shows a color bar
    
    keyword args:
     cmap: color map
     origin: 'lower' puts 0,0 at the lower left corner
             'center' puts 0,0 at the lower left of pixel n/2 (pixels are numbered 0 to n-1)
     dx: increment of one pixel
    """
    assert type(z) == np.ndarray
    assert len(z.shape) == 2

    n,m = z.shape
    extent = np.array([0,n,0,n])
    if (origin == 'center'):
        extent = extent - n/2
    extent *= dx

    def format_coord(xx, yy):
        col = int(xx+0.5)
        row = int(yy+0.5)
        l,r,b,t = extent
        row = int(((yy-b)/(t-b))*m)
        col = int(((xx-l)/(r-l))*n)
        if col>=0 and col<n and row>=0 and row<m:
            return '%1.2f,%1.2f[%1.0f,%1.0f] %1.2f' % (xx,yy, row,col, z[row,col])
        else:
            return 'x=%1.4f, y=%1.4f' % (xx, yy)
    
    plt.ion()
    plt.figure()
    ax = plt.subplot(111)
    im = ax.imshow(z,interpolation='nearest',origin='lower',extent=extent, cmap=cmap)
    ax.format_coord = format_coord
    plt.colorbar(im)
    if (grid):
        plt.grid('on')
    return im
    
def dimg(z,fig=None,sub=None,title=None,ap=None,extent=None,stride=None,image=True,contour=False,levels=None,surface=False,cmap='bone',origin='upper',colorbar=False,geometry=None):
    """Displays an array graphically.
    
    Usage: ret = dimg(theArray,[options])
    options:
        fig (None) - the existing figure window to draw in
        sub (None) - the subplot region to draw in, if any
        image - create a grey-scale image of the 2d data (default)
        contour - draw contour lines. can be combined with image
        surface - create a 3d surface rendering
        cmap ('bone') can be 'hot'
    returns:
        image object, contour object, (image,contour), or surface object
        depending on the option choices
    """
    if surface:
        image = False
        contour = False
    
    if contour:
        origin = 'lower'
    
    if (type(z) == list):
        movie(z,ap=ap)
        return
    
    if (machine == 'rtc.ucolick.org'):
        print '<dimg> not available on ',machine
        print '<dimg> use dimg.movie([arg]) instead to force display using ds9'
        return
    
    assert isinstance(z,np.ndarray),'<dimg> ERROR argument must be a 2d array'
    
    if (len(z.shape) == 3):
        movie(z,ap=ap)
        return
    
    if (ap != None):
        z = z*ap
    
    plt.ion()
    
    try:
        thefig = plt.figure(fig)
    except:
        if (type(fig) == type('string')):
            if (title is None):
                title = fig
            fig = None
    
        thefig = plt.figure(fig)

    if (fig is not None) and (sub is None):
        plt.clf()
    
    if (sub is not None):
        ax = plt.subplot(sub)
    else:
        ax = plt.subplot(111)

    if (title is not None):
        plt.title(str(title))    
    
    def format_coord(xx, yy):
        col = int(xx+0.5)
        row = int(yy+0.5)
        if extent is not None:
            l,r,b,t = extent
            row = int(((yy-b)/(t-b))*numrows)
            col = int(((xx-l)/(r-l))*numcols)
            if (origin != 'lower'):
                row = numrows-row-1
        if col>=0 and col<numcols and row>=0 and row<numrows:
            #return 'x%1.2fy%1.2f[r%1.0fc%1.0f]v%1.2f'%(xx,yy, row,col, z[row,col])
            return 'x%1.2f y%1.2f [r%1.0fc%1.0f] z%1.2f'%(xx,yy, row,col, z[row,col])
        else:
            return 'x=%1.4f, y=%1.4f'%(xx, yy)

    numrows, numcols = z.shape
    if extent is None:
        x = range(numcols)
        y = range(numrows)
    else:
        x = np.linspace(extent[0],extent[1],numcols)
        y = np.linspace(extent[2],extent[3],numrows)
    
    im = None
    if image:
        im = ax.imshow(z,extent=extent,cmap=cmap,interpolation='nearest',origin=origin)
        ax.format_coord = format_coord
        if (colorbar):
            plt.colorbar(im)
        ret = im
    if contour:
        if levels is None:
            if image:
                cs = plt.contour(z,extent=extent,colors='lightgreen')
            else:
                cs = plt.contour(z,extent=extent,colors='black')
        else:
            if image:
                colors = None
            else:
                colors = 'black'
                ax.format_coord = format_coord
            cs = plt.contour(z,extent=extent,levels=levels,colors=colors)
            plt.clabel(cs)
        plt.clabel(cs)
        if image:
            ret = (im,cs)
        else:
            ret = cs
    if surface:
        # set the stride to keep from plotting too many points, unless the stride is overridden
        surface_maxpts = 20
        if stride is None:
            n,m = z.shape
            stride = np.array([n,m])/surface_maxpts
            stride = np.clip(stride,1,np.inf).astype(int)
        ax = plt.gca(projection='3d')
        x2,y2 = np.meshgrid(x,y)
        z2 = z.copy()
        if (isinstance(z,np.ma.core.MaskedArray)):
            x2 = np.ma.array(x2,mask=z.mask).compressed()
            y2 = np.ma.array(y2,mask=z.mask).compressed()
            z2 = z2.compressed()
            surf = ax.plot_trisurf(x2,y2,z2,cmap=cmap)
        else:
            surf = ax.plot_surface(x2,y2, z2, rstride = stride[0], cstride = stride[1], cmap=cmap)
        ret = surf

    if geometry is not None:
        pass
        #mgr = plt.get_current_fig_manager()
        #mgr.window.geometry(geometry)
    
    def _onclick(event):
        #print 'figure = '+thefig.__repr__()
        #print 'button=%d, x=%d, y=%d, xdata=%f, ydata=%f'%(
        #    event.button, event.x, event.y, event.xdata, event.ydata)
        #print 'xlimits: '+str(plt.xlim())+'  ylimits: '+str(plt.ylim())
        if (not image):
            return
        toolstate = thefig.canvas.manager.toolbar._active
        if (toolstate is None and event.button == 3):
            z_range = np.array([z.min(),z.max()])
            x_range = plt.xlim()
            y_range = plt.ylim()
            brightness = (event.ydata - y_range[0])/(y_range[1]-y_range[0])
            contrast = (event.xdata - x_range[0])/(x_range[1]-x_range[0])
            dz = z_range[1]-z_range[0]
            mid = (z_range[0]+z_range[1])/2.
            z1 = z_range[1] - ((brightness - 0.5)/0.5)*dz
            dz = dz*contrast*4
            z0 = z1 - dz
            #print 'brightness = '+str(100*brightness)+' %'
            #print 'contrast = '+str(100*contrast)+' %'
            ax = plt.subplot(111)
            ax.imshow(z,vmin=z0,vmax=z1,extent=extent,cmap=cmap,interpolation='nearest',origin=origin)

    cid = thefig.canvas.mpl_connect('button_press_event', _onclick)
    return ret

show = dimg

def ds9(arr2d,ap=None):
    """
    Displays a 2-d array using ds9
    """
    movie([arr2d],ap)

    
def movie(arr3d,ap=None):
    """
    Displays a 3-d data cube using ds9
    The array can be a 3-dimensional array or a list of 2-d arrays
    ap is an optional aperture window outside of which values are not displayed
    """
    if (type(arr3d) == list):
        movie(np.array(arr3d),ap=ap)
        return
    if (type(arr3d) != np.ndarray):
        print '<movie> ERROR argument must be an array'
        return
    if (len(arr3d.shape) != 3):
        print '<movie> ERROR array argument must have 3 dimensions'
        return
    if (ap != None):
        for i in range(arr3d.shape[0]):
            arr3d[i,:,:] *= ap
    os.system("rm -f temp_ds9_array.fits")
    fits.writefits(arr3d,'temp_ds9_array.fits')
    os.system("ds9 temp_ds9_array.fits &")

def play(arr3d,dt_ms=50,ap=None,cmap='bone',label = 'time',times=None,tunits='',**kwargs):
    """play(arr3d,ap=None,rate=1000,loop=False)
    Displays a 3-d data cube using matplotlib
    The array can be a 3-dimensional numpy array or a list of 2-d numpy arrays
    ap is an optional aperture window outside of which values are not displayed
    """
    #global k,frames,ani,pause, step
    
    arr3d = np.array(arr3d)
    if ap is not None:
        arr3d = arr3d*ap
    
    vmin = arr3d.min()
    vmax = arr3d.max()
    
    shape = arr3d.shape
    if (len(shape) == 3):
        nt,nr,nc = shape
        frames = arr3d
    if (len(shape) == 4): # this is a block display
        na,nt,nr,nc = shape
        frames = arr3d.transpose([1,2,0,3]).reshape((nt,nr,na*nc))
    if (len(shape) == 5):
        na,nb,nt,nr,nc = shape
        frames = arr3d.transpose([2,0,3,1,4]).reshape((nt,na*nr,nb*nc))

    numrows, numcols = frames[0].shape
    
    fig = plt.figure()
    ax = fig.gca()

    im = plt.imshow(frames[0],vmin=vmin,vmax=vmax,cmap=cmap,interpolation='nearest',animated=True,**kwargs)
    n = len(frames)
    if times is None:
        times = np.arange(float(n))
    tstr = '{:8.4f}'.format(float(times[0]))
    plt.title(label+' '+tstr+' '+tunits)
    
    k = 0
    z = frames[0]
    n = len(frames)
    step = +1
    
    def updatefig(*args):
        #global k
        #print '<updatefig> ---------------'
        #print args
        c = args[1]
        fig = c.figure
        plt.figure(fig.number)
        #print c
        #print '---------------------------'

        z = c.frames[c.k]
        im.set_array(z)
        tstr = '{:8.4f}'.format(float(times[c.k]))
        plt.title(label+' '+tstr+' '+tunits)
        if not c.pause:
            if c.report_time:
                if (c.k == 0):
                    c.t0 = time.time()
                if (c.k+c.step) >= n:
                    t1 = time.time()
                    print 'display rate:',float(n)/(t1-c.t0),'Hz'
                    c.report_time = False
            c.k = (c.k+c.step) % n
        return im,
    
    def onClick(event):
        #global pause
        c = event.canvas
        c.pause ^= True
    
    def onKey(event):
        #global pause, step
        #print event.key
        c = event.canvas
        if event.key == 'escape':
            c.frames = 0
            plt.close()
        if event.key == 'right':
            c.pause = False
            c.step = +1
            updatefig(0,c)
            c.pause = True
        if event.key == 'left':
            c.pause = False
            c.step = -1
            updatefig(0,c)
            c.pause = True
            c.step = +1
        if event.key == ' ':
            c.pause ^= True        
    
    pause = False
    fig.canvas.mpl_connect('button_press_event',onClick)
    fig.canvas.mpl_connect('key_press_event',onKey)
    fig.canvas.k = k
    fig.canvas.frames = frames
    fig.canvas.pause = pause
    fig.canvas.step = step
    fig.canvas.report_time = True
    fig.canvas.t0 = 0.
    ani = animation.FuncAnimation(fig,updatefig,fargs = [fig.canvas],interval=33,blit=False)
    fig.canvas.ani = ani
    plt.show()
    print 'instructions:'
    print '  space or left mouse click: pause'
    print '  right arrow: advance one frame forward'
    print '  left arrow: go back one frame'
    print '  escape: exit'

def play_test():
    global frames
    frames = np.random.normal(size=(100,100,100))
    play(frames)

def clearwins():
    """Close all the dimg display windows
    """
    plt.close('all')

def example(surface=False):
    """Run a simple example with generated dummy data
    """
    global x,y,z,extent
    
    npts = 100
    stride = [1,1]
    if surface:
        stride = np.fix(np.array([1,1])*(npts/20.)).astype('int')
    x = np.linspace(-2*np.pi,2*np.pi,npts)
    x,y = np.meshgrid(x,x)
    z = x**2*y**3
    extent = np.array([-1,1,-1,1])*2*np.pi
    dimg(z,origin='lower',extent=extent,surface=surface,stride=stride,colorbar=True)

def test():
    """Regression-test the dimg package
    """
    f = '(x^2)*(y^2)'
    example(); plt.title('image '+f)
    example(surface=True); plt.title('surface '+f)
    levels = np.logspace(0,5,num=6)
    levels = np.hstack([-levels[::-1],0,levels])
    img,cs = dimg(z,contour=True,origin='lower',extent=extent,levels=levels); plt.title('image & contour '+f)
    cs = dimg(z,image=False,contour=True,extent=extent,levels=levels); plt.title('contour '+f)
    plt.grid('on')
    
__version__ = "1.2 Aug 8, 2015"

if __name__ == '__main__':
    test()
    plt.show(block=False)
    u = raw_input('hit return to exit')
    print 'done'



# read and display a fits file
import os
import sys
import numpy as np
import pyfits
import time
import warnings

def readfits(filename,verbose=False,raiseError=True,isHST=False,return_header=False):
    """Reads a FITS file from disk given the file name."""
    if (verbose):
        print '<readfits.readfits> reading file '+filename + '...',
        sys.stdout.flush()
    if (os.path.isfile(filename) == False):
        err_str = "<readfits.readfits> no such file "+filename
        if (raiseError):
            raise IOError(err_str)
            return
        warnings.warn(err_str)
        return np.array([0.])
    hdulist = pyfits.open(filename)
    if (isHST):
        data = hdulist[1].data
        hdr = hdulist[1].header
    else:
        data = hdulist[0].data
        hdr = hdulist[0].header
    hdulist.close()
    if (verbose):
        print 'done'
    if (return_header):
        return data,hdr
    return data

def writefits(data,filename,header=None,protect=False,comment=None,verbose=False):
    """Writes a FITS file given the data.
    arguments:
        data - a numpy array (1, 2 or 3d)
        filename - full name of the file to write to
    keywords:
        hdr - a list of tuples that will be appended as header entries
            these are (keyword,value,[comment]) tuples
        protect - avoid clobering an existing file with the same name
        comment - a single comment to append to the header
        verbose - print progress to the terminal
    """

    if (os.path.isfile(filename) == True):
        if (protect):
            protectFile(filename,verbose=verbose)
        else:
            err_str = "<readfits.writefits> won't over write existing file "+filename + ". Try the protect=True keyword."
            raise IOError(err_str)
            return
    if (verbose):
        print '<readfits.writefits> writing file '+filename + '...',
        sys.stdout.flush()
    tl = time.localtime(time.time())
    day = time.strftime('%Y-%m-%d',tl)
    tim = time.strftime('%H%M%S',tl)
    hdu = pyfits.PrimaryHDU(data)
    if (isinstance(header,pyfits.header.Header)):
        hdu.header = header
    if (isinstance(header,list)):
        for h in header:
            hdu.header.append(h,end=True)
    hdu.header.append(('COMMENT','------ fits writer info -------'),end=True)
    hdu.header.append(('DATE',day,'File generation date'),end=True)
    hdu.header.append(('TIME',tim,'file generation time hhmmss'),end=True)
    hdu.header.append(('GENER','readfits.py','writer code'),end=True)
        
    if (comment is not None):
        hdu.header.append(('comment',str(comment)),end=True)
    hdu.writeto(filename)
    if (verbose):
        print 'done'

def protectFile(filename,verbose=False):
    """Protect an existing file by putting it into an
    archive subdirectory and appending a datetime stamp to the name
    """
    if (os.path.isfile(filename)):
        dirname = os.path.dirname(filename)
        basename = os.path.basename(filename)
        archivedir = os.path.join(dirname,'archive')
        if (not os.path.exists(archivedir)):
            os.mkdir(archivedir)
        protectedName = os.path.join(archivedir,basename)+'.'+time.strftime('%Y%m%d%H%M%S')
        if (verbose):
            print '<readfits.protectFile> protecting file '+filename + ' to '+protectedName,
            sys.stdout.flush()            
        os.rename(filename,protectedName)
        if (verbose):
            print 'done'



"""Front end (test)
"""
import os
import sys
import time
flush = sys.stdout.flush
import ConfigParser as cp
import psf as p
import numpy as np
import infoArray
import warnings
import platform
import matplotlib.pylab as plt
from matplotlib.backends.backend_pdf import PdfPages

def run(configFile=None,verbose=True):
    global a
    
    #------- default arguments --------
    defaults={
        'ngs':3,
        'geom':'circle',
        'rot': '45.',
        'radius':'6 arcmin',
        'profile':'MaunaKea 30 N',
        'outer_scale':'30 meters',
        'r0': '',
        'DM_conjugate': '-280 meters',
        'actuator_spacing': '0 cm',
        'wavelength':'0.5 microns',
        'field_points': '0., 2.5, 4., 5.',
        'filename':'psf.fits',
        'seeing_psf_filename':None
    }
    
    t0 = time.time()
    #-------- read arguments from config file ------
    if verbose:
        print 'reading config params...',; flush()
    path = os.path.abspath(__file__)
    dir_path = os.path.dirname(path)
    exampleConfigFile = os.path.join(dir_path,'example.cfg')

    if configFile is None:
        configFile = exampleConfigFile
    
    if not os.path.isfile(configFile):
        if configFile == 'example.cfg':
            configFile = exampleConfigFile
        else:
            raise IOError,'no file named %s'%configFile

    if configFile == exampleConfigFile:
        print ''
        print 'WARNING using example config file from code source directory'
        print '%s'%exampleConfigFile
            
    config  = cp.ConfigParser(defaults)
    config.read(configFile)
    
    ngs = config.get('constellation','ngs')
    geom = config.get('constellation','geom')
    radius = config.get('constellation','radius')
    rot = config.get('constellation','rot')
    profile = config.get('atmosphere','profile')
    outer_scale = config.get('atmosphere','outer_scale')
    r0 = config.get('atmosphere','r0')
    zc = config.get('AO','DM_conjugate')
    actSpacing = config.get('AO','actuator_spacing')
    wavelength = config.get('image','wavelength')
    fieldpts = config.get('image','field_points')
    filename = config.get('output','filename')
    seeing_psf_filename = config.get('output','seeing_psf_filename')
    
    #---------- parse and interpret arguments -------
    ngs = int(ngs)
    r= radius.split()
    radius = float(r[0])
    if len(r) > 1:
        ufac = {'arcmin':1,'arcsec':60}[r[1]]
        radius *= ufac
    r= outer_scale.split()
    outer_scale = float(r[0])
    if len(r) > 1:
        ufac = {'meter':1,'meters':1,'m':1,'cm':0.01}[r[1]]
        outer_scale *= ufac
    r= wavelength.split()
    wavelength = float(r[0])
    if len(r) > 1:
        ufac = {'micron':1,'microns':1,'angstrom':0.1,'angstroms':0.1}[r[1]]
        wavelength *= ufac
    r= zc.split()
    zc = float(r[0]) # only units of meters is allowed
    r= fieldpts.split(',')
    fieldpts = map(float,r)
    rot = float(rot)
    r= actSpacing.split()
    actSpacing= float(r[0])
    if len(r) > 1:
        ufac= {'m':1,'meter':1,'meters':1,'cm':0.01}[r[1]]
        actSpacing *= ufac
    if actSpacing < 0.01:
        actSpacing = None # spacing less than a centimeter => 'infinite' sampling
    
    valid_profiles = p.available(show=False)
    assert profile in valid_profiles, 'profile must be one of %r'%valid_profiles
    
    t1 = time.time()
    if verbose:
        print '(%.3f sec)'%(t1-t0); t0=t1
    #---------- do PSF calculation -------
        print 'doing PSF calculation...',; flush()
    N = ngs
    a = p.Cn2_Profile(profile,L0=outer_scale,lam=wavelength,sf_method='new')
    n = 256
    field_point = [fieldpts[0],0.]
    c = p.Constellation(N,radius,'circle',rot=rot,field_point = field_point)
    r_lim = a.r0*n/8.
    a.make_S(c,r_lim=r_lim,n=n,dm_conjugate=zc,wfs_conjugate=zc, act_spacing = actSpacing)
    w = np.zeros(N+1)
    w[0] = -1
    w[1:] = np.ones(N)/float(N)
    a.make_PSF(w)
    
    t1 = time.time()
    if verbose:
        print '(%.3f sec)'%(t1-t0); t0=t1
    #---------- save result --------
        print 'saving result...',; flush()
    keys_a = ['name','site','tile','Cn2','Cn2_units','Cn2_bar',
            'databaseFile','databaseMetadata',
            'r0','r0_units','r00','r00_units','spatialFilter',
            'theta0','theta0_units',
            'L0','L0_units',
            'dm_conjugate','dm_conjugate_units',
            'wfs_conjugate','wfs_conjugate_units']
    transkey = {'name':'profile','databaseFile':'dbFile','databaseMetadata':'dbMeta'}
    keys_c = ['N','field_point','geometry','radius','radius_units',
              'rotation','rotation_units']
    
    if a.spatialFilter is not False:
        keys_a += ['act_spacing']
    
    hdu = a.PSF._hdu()
    hdr = hdu.header
    hdr.append(('COMMENT','--- Guide Star Constellation ---'),end=True)
    d = a.constellation.__dict__
    for key in keys_c:
        if not key.endswith('_units'):
            kwd = key.upper()[:8]
            val = d[key]
            if isinstance(val,(tuple,list,np.ndarray)):
                val = str(str(val))
            if isinstance(val,dict):
                val = str(val)
            if isinstance(val,float):
                val = ['Infinity',val][np.isfinite(val)]
            card = (kwd,val)
            if key+'_units' in keys_c:
                cmt = d[key+'_units']
                card = card + (cmt,)
            hdr.append(card,end=True)
    d = a.__dict__
    hdr.append(('COMMENT','--- Atmosphere characteristics ---'),end=True)
    for key in keys_a:
        if not key.endswith('_units'):
            if key in transkey:
                kwd = transkey[key]
            else:
                kwd = key.upper()[:8]
            val = d[key]
            if key == 'spatialFilter' and not isinstance(val,bool):
                val = True
            if isinstance(val,(tuple,list,np.ndarray)):
                val = str(list(val))
            if isinstance(val,dict):
                val = str(val)
            if isinstance(val,float):
                val = ['Infinity',val][np.isfinite(val)]
            card = (kwd,val)
            if key+'_units' in keys_a:
                cmt = d[key+'_units']
                card = card + (cmt,)
            hdr.append(card,end=True)
    
    
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore',module='pyfits')
        hdu.writeto(filename,clobber=True)
    
    a.PSF_seeing.profile = a.profile
    a.PSF_seeing.Cn2 = str(list(a.Cn2))
    a.PSF_seeing.L0 = ['Infinity',a.L0][np.isfinite(a.L0)]
    a.PSF_seeing.L0_units = a.L0_units
    hdu = a.PSF_seeing._hdu()
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore',module='pyfits')
        hdu.writeto(seeing_psf_filename,clobber=True)
    
    t1 = time.time()
    if verbose:
        print '(%.3f sec)'%(t1-t0); t0=t1
        print 'done'; flush()

def summary(pdf=False):
    print 'GLAO PSF ENA = %0.4f %s'%(a.PSF.ena,a.PSF.ena_units)
    print 'Seeing PSF ENA = %0.4f %s'%(a.PSF_seeing.ena,a.PSF_seeing.ena_units)
    print 'GLAO PSF ENA_r = %0.4f %s'%(a.PSF.ena_r,'arcsec')
    print 'Seeing PSF ENA_r = %0.4f %s'%(a.PSF_seeing.ena_r,'arcsec')
    assert pdf in [True,False]
    if platform.system() == 'Darwin' and 'VIRTUAL_ENV' in os.environ:
        pdf = True
    if pdf:
        gfile = 'glao_psf_summary.pdf'
        pp = PdfPages(gfile)
    plt.figure()
    a.constellation.graph()
    plt.grid('off')
    if pdf:
        pp.savefig()
    a.PSF.graph(label='GLAO')
    a.PSF_seeing.graph(oplot=True,label='seeing')
    plt.legend()
    if pdf:
        pp.savefig()
    a.PSF.show()
    if pdf:
        pp.savefig()
    a.PSF_seeing.show()
    if pdf:
        pp.savefig()
        pp.close()
        print 'Graphs written to %s'%gfile

if __name__ == 'main':
    if len(sys.argv >1):
        configFile = sys.argv[1]
    else:
        configFile = 'example.cfg'
    run(configFile)



"""
img.py

Image processing routines
"""
import numpy as np

try:
    import scipy.special
    import scipy.interpolate
    import scipy.ndimage
    from scipy.special import jn
    no_scipy = False
except ImportError:
    no_scipy = True # but proceed anyway - some functions are disabled

def circle(shape,c=None,r=None):
    """circle(shape,c=None,r=None)
    Generate a 2-d image with a disk of ones inside radius r and centered at c, zero outside
    
    c = a tuple of the x,y center position of the center of the circle, in pixels
    r = a scalar radius of the circle, in pixels
    """
    if (c is None):
        c = np.array(shape)/2.
    y = np.arange(shape[0]).astype('float') - float(c[0])
    x = np.arange(shape[1]).astype('float') - float(c[1])

    if (r is None):
        r = min(np.array(shape))/2.
    x = x/float(r)
    y = y/float(r)
    x,y = np.meshgrid(x,y)
    r = np.sqrt(x**2+y**2)
    result = np.where(r<1,1.,0.)
    return result

def supergauss(shape,rad,c=None,pow=10):
    """Generate a super-gaussian mask:
    e^(-r^10)
    """
    x1 = np.arange(shape[0])
    y1 = np.arange(shape[1])
    x,y = np.meshgrid(x1,y1)
    if (c is None):
        c = (shape[0]/2,shape[1]/2)
    x -= c[0]
    y -= c[1]
    r = np.sqrt(x**2+y**2)/rad
    u = np.exp(-r**pow)
    return u

def gauss2(shape,s,c=None):
    """gauss2(shape,sd,c=None)
    Generate a Gaussian shaped 2-d bump, centered at c and with width (std dev) sd
    
    c = a tuple of the x,y center position of the Gaussian bump, in pixels
    sd = the standard deviation (1/e point) of the Gaussian bump, in pixels
        - sd can be a scalar, in which case it applies to both axes with corellation coeficent zero
        -or- a tuple of 2 elements in which case it indicates widths in each axis woth corellation coefficient zero
        -or- a tuple of 3 elements in which case it represents the elliptical orientation:
            standard deviations in the major axes and the rotation angle in radians
        -or- a 2x2 matrix, in which case it represents the Information matrix
    """
    if (c is None):
        c = np.array(shape)/2.
    x = np.arange(shape[0]).astype('float') - float(c[0])
    y = np.arange(shape[1]).astype('float') - float(c[1])

    x,y = np.meshgrid(x,y)
    F = None
    
    type_s = str(type(s))
    valid_matrix_types = ["<type 'numpy.ndarray'>","<class 'numpy.core.defmatrix.matrix'>","<class 'numpy.matrixlib.defmatrix.matrix'>"]
    if (type(s) in (tuple,list)):
        sx = float(s[0])
        sy = float(s[1])
        if (len(s) == 2):
            theta = 0.
        else:
            theta = s[2]
    elif (type_s in valid_matrix_types):
        if (s.shape == (2,2)):
            F = np.matrix(s)
        else:
            return None
    else: # assume its a scalar
        sx = float(s)
        sy = float(s)
        theta = 0.
    
    # now form the Information matrix
    if (F is None):
        C = np.matrix(np.zeros((2,2)))
        C[0,0] = sx**2
        C[1,1] = sy**2
        R = np.matrix([[np.cos(theta),-np.sin(theta)],[np.sin(theta),np.cos(theta)]])
        F = (R.T*C*R).I
    
    r2 = F[0,0]*x**2 + 2*F[0,1]*x*y + F[1,1]*y**2
    result = np.exp(-r2)
    return result

def circle_backup(shape,center,radius):
    n,m = shape
    x = range(n)
    y = range(m)
    xx,yy = np.meshgrid(x,y,indexing='ij')
    r = np.sqrt((xx-center[0])**2 + (yy-center[1])**2)
    result = (r < radius).astype(float)
    return result

def depiston(image,ap=None):
    """Remove the piston component, or piston component
    within an aperture, from an image
    """
    assert image.ndim >= 2
    if (ap is None):
        ap = np.ones(n,m)
    if image.ndim == 2:
        n,m = image.shape
        piston = np.sum(image*ap)/np.sum(ap)
        result = image - piston
        return result
    else:
        s = image.shape
        ni = np.product(s[:-2])
        image = image.reshape(ni,s[-2],s[-1])
        piston = np.sum(image*ap,axis=0)/np.sum(ap)
        result = image - piston
        result = result.reshape(s)
        return result

def detilt(image,ap=None):
    """Remove the tilt component, or the tilt component
    on an aperture, from an image
    """
    n,m = image.shape
    if (ap is None):
        ap = np.ones(n,m)
    x = np.arange(n)-n/2
    y = np.arange(m)-m/2
    xx,yy = np.meshgrid(x,y)
    xx = xx - np.sum(xx*ap)/np.sum(ap)
    yy = yy - np.sum(yy*ap)/np.sum(ap)
    tx = np.sum(xx*image*ap)/np.sum(xx*xx*ap)
    ty = np.sum(yy*image*ap)/np.sum(yy*yy*ap)
    result = image - (xx*tx + yy*ty)
    return result

def planeFit(image,ap=None):
    """Determine the plane that fits the image, over an aperture
    Returns the plane fit and the piston, tilt-x, and tilt-y components, in a tuple
    """
    n,m = image.shape
    if (ap is None):
        ap = np.ones(n,m)
    piston = np.sum(image*ap)/np.sum(ap)
    x = np.arange(n)-n/2
    y = np.arange(m)-m/2
    xx,yy = np.meshgrid(x,y)
    xx = xx - np.sum(xx*ap)/np.sum(ap)
    yy = yy - np.sum(yy*ap)/np.sum(ap)
    tx = np.sum(xx*(image-piston)*ap)/np.sum(xx*xx*ap)
    ty = np.sum(yy*(image-piston)*ap)/np.sum(yy*yy*ap)
    p = (piston + xx*tx + yy*ty)
    return (p,piston,tx,ty)

def rms(image,ap):
    """Computes the root-mean-square of an image, over an aperture
    """
    n,m = image.shape
    if (ap is None):
        ap = np.ones(n,m)
    u = depiston(image,ap)
    result = np.sqrt(np.sum(u*u*ap)/np.sum(ap))
    return result

def zeropad(u,n=None,pad=None):
    """
    r = zeropad(u,n)
    
    Zero-pad the array by centering it in a larger
    array and padding with zeros
    
    u - original array
    n - size of new array (2-tuple)
    pad - added size of new array, so size of new array is size of
           the original array with a 'pad' size buffer all around
    """
    n0 = np.array(u.shape)
    if n is not None:
        n = np.array(n)
    if pad is not None:
        n = u.shape + 2*np.array(pad)
    broadcast = False
    if (len(n0) > 2):
        broadcast = True
        n0 = n0[-2:]
    assert (n0<=n).all()
    d = (n-n0)/2
    d2 = (n-n0) - d
    p = ((d2[0],d[0]),(d2[1],d[1]))
    if broadcast:
        r = []
        for uu in u:
            r.append(np.pad(uu,p,'constant',constant_values=0))
        r = np.array(r)
    else:
        r = np.pad(u,p,'constant',constant_values=0)
    return r

def shift(im,s):
    """
    shift the 1-d or 2-d image by the amounts in the s tuple,
    s[0] is the shift in 'x' (columns) and s[1] is the shift in 'y' (rows)
    """
    if (isinstance(s, int)):
        s = (s,)
    ndim = len(s)
    assert (ndim > 0 and ndim <= 2)
    if (ndim == 1):
        r = np.roll(im,s[0])
    if (ndim == 2):
        r = np.roll(im,s[0],axis=1)
        r = np.roll(r,s[1],axis=0)
    return r

def blur(a,p,kernelType='block'):
    """blur an image by a PSF or by a kernel block
    kernelType can be 'block', 'circle', 'Gaussian' or 'Airy'
    
    if kernelType is 'PSF' then p must be an image
    and a convolution is performed
    if kernelType is 'block' then p must be an
    integer, and the kernel is then a pxp block
    if kernelType is 'circle' then p
    represents the diameter of a circular kernel
    if kernelType is 'Gaussian' or 'Airy' then the
    appropriate kernel is generated, with p representing
    in the Gaussian case standard deviation,
    and in the Airy case, the distance from center to
    first Airy null.
    'block', 'Gaussian', and 'Airy' blur functions
    are normalized to preserve flux
    """
    n,m = a.shape[-2:]
    assert kernelType in ['PSF','block','Gaussian','Airy','circle']
    if (kernelType == 'PSF'):
        b = p
        nb,mb = b.shape
        assert (nb<=n and mb<=m)
    if (kernelType == 'block'):
        assert isinstance(p,int)
        nb,mb = p,p
        b = np.ones((nb,mb))
        b = b/np.sum(b)
    if (kernelType == 'circle'):
        b = circle((n,m),(n/2,m/2),p)
        b = b/np.sum(b)
    if (kernelType == 'Gaussian'):
        b = gauss2((n,m),p,(n/2,m/2))
        b = b/np.sum(b)
        nb,mb = n,m
    if (kernelType == 'Airy'):
        b = airy(n,m,p)
        b = b/np.sum(b)
        nb,mb = n,m
    c = convolve(a,b)
    return c

def convolve(a,b,is_complex=False):
    """convolve data a with kernel b
    using the Fourier transfomrm method
    Assumes data sets are 2d arrays and (0,0)
    is at the center of the array
    """
    n,m = a.shape[-2:]
    nb,mb = b.shape
    assert nb<=n and mb<=m
    if (nb<n or mb<m):
        b = zeropad(b,(n,m))
    a_f = np.fft.fft2(a)
    b = np.fft.ifftshift(b)
    b_f = np.fft.fft2(b)
    c_f = a_f*b_f
    c = np.fft.ifft2(c_f)#/(float(n)*float(m))
    if not is_complex:
        c = np.real(c)
    return c

def airy(n,m,s,s2=None):
    """
    PSF of a circular aperture
    args:
        n,m = size of array
        s = distance in pixels to the first Airy null
    keyword args:
        s2 = secondary Airy null - used for the case of a secondary obscuration
        
    returns:
        the (annular) airy function, computed analytically, and normalized to peak = 1.
        
    examples:
    
        p = airy(n,n,1.22*(lambda/d)/pixelscale)
        p2 = airy(n,n,1.22*(lamda/dp)/pixelscale, s2 = 1.22*(lambda/ds)/pixelscale)
    """
    xx = np.arange(n) - n/2
    yy = np.arange(m) - m/2
    x,y = np.meshgrid(xx,yy)
    r = np.sqrt(x**2+y**2)
    r = np.clip(r,1.e-12,r.max())
    a = 1.219669891266504 # first zero of j1 is at a*pi
    f = jn(1,a*r/s*np.pi)/r
    norm_factor = 2./(a*np.pi/s) # normalizes to peak = 1.0
    if s2 is not None:
        f -= (s/s2)*jn(1,a*r/s2*np.pi)/r # to subtract the secondary diffraction, it must be normalized by area
        norm_factor *= 1./((s/s2)**2-1.) # and the peak-normalzing factor adjusted
    f *= norm_factor
    ret = f**2
    return ret

def ft(a):
    """2-D centered fast-Fourier transform
    """
    return np.fft.fftshift(np.fft.fft2(np.fft.fftshift(a)))

def ftinv(a):
    """2-D centered inverse fast-Fourier transform
    """
    return np.fft.ifftshift(np.fft.ifft2(np.fft.ifftshift(a)))

def ftconvolve(a,b,space='sss'):
    """2-D centered fast-Fourier convolve
    
    result = a (*) b
    space - tell the domain of the data or result, 's' = spatial, 'f' = fourier
        It must be 3 characters, refering to a,b, and result
        For exammple: space = 'fss' says a is already transformed, b is not,
        and the returned result should be in the spatial domain
    """
    if space[0] == 's':
        fa = ft(a)
    else:
        fa = a
    if space[1] == 's':
        fb = ft(b)
    else:
        fb = b
    fc = fa*fb
    if space[2] == 's':
        return ftinv(fc)
    else:
        return fc

class units:
    arcsec = np.pi/(180.*3600.)
    nm = 1.e-9 
    micron = 1.e-6 

def nextpow2(n):
    k = 0
    n2 = 2**k
    while n2 < n:
        k += 1
        n2 = 2**k
    return n2

def psfwf(mag,phase,pad=0):
    """compute the PSF for an aberrated wavefront
    given the magnitude (or aperture) and phase
    """
    n,m = phase.shape
    assert (n,m) == mag.shape
    
    if (pad > 0):
        nn = nextpow2(n*2**pad)
    else:
        nn = n
    
    wf = mag*np.exp(1j*phase)
    wf = zeropad(wf,(nn,nn))
    wf = np.fft.fftshift(wf)
    p = np.fft.fft2(wf)/float(nn)
    p = np.fft.fftshift(p)
    if (n < nn):
        p = crop(p,(nn/2,nn/2),(n,n))
    return np.abs(p)**2    

def psf(n,dp,ds,lam,pixelscale,pad=0,center=None):
    """
    PSF of a circular aperture with a secondary obscuration
    
    args:
        n,n = size of array
        dp, ds = diameter of primary and secondary in meters
        lam = wavelength, meters
        pixelscale = size of pixel, in radians
    
    keyword args:
     pad: pad in the aperture domain, which reduces the 'Fourier noise' in the focal domain result,
        at the cost of requiring more compute time to calculate.
        The pad factor is the exponent of two in the padding, i.e. pad=1 implies pad the array
        by a factor of 2, pad=3 pads by a factor of 8, etc.
     center: center the psf at this pixel. can be a fractional pixel. pixels are numbered
        0 to (n-1). pixel k's continuous coordinate range is k to k+1.
        the default center is n/2+0.5, which is the center of pixel n/2

    returns:
        the PSF, normalized so integral = 1.
    
    """
    if (pad > 0): 
        nn = nextpow2(n*2**pad)
    else:
        nn = n
    c0 = np.array([n/2+0.5,n/2+0.5])
    if (center is None):
        center = c0
    s = np.array(center) - c0
    
    # create a pupil
    dx = pixelscale
    du = lam/(nn*dx)
    ap = circle((nn,nn),(nn/2,nn/2),dp/du/2.)
    if (ds > 0):
        ap -= circle((nn,nn),(nn/2,nn/2),ds/du/2.)
    mag = ap/np.sqrt(np.sum(ap*ap))
    
    # incorporate a phase factor on the aperture to do the fractional pixel shift
    u = np.arange(0,nn) - nn/2
    u,v = np.meshgrid(u,u)
    ph = 2*np.pi * (s[0]*u + s[1]*v) * ap / float(nn)
    wf = mag * np.exp(1j*ph)
    
    # transform to get the PSF
    wf = np.fft.fftshift(wf)
    p = np.fft.fft2(wf)/float(nn)
    p = np.fft.fftshift(p)
    if (n < nn):
        p = crop(p,(nn/2,nn/2),(n,n))
    return np.abs(p)**2

def crop(z,center,size,fill=0.):
    """crop a 2D image to size
    """
    n,m = size
    nn,mm = z.shape
    w = np.zeros((n,m)) + fill
    ll = center[0] - n/2
    rr = ll + n
    bb = center[1] - m/2
    tt = bb + m
    l = 0
    r = n
    b = 0
    t = m
    if ll < 0:
        d = abs(ll)
        l += d
        ll += d
    if rr >= nn:
        d = rr - nn + 1
        r -= d
        rr -= d
    if bb < 0:
        d = abs(bb)
        b += d
        bb += d
    if tt >= mm:
        d = tt - mm + 1
        t -= d
        tt -= d
    w[ l:r, b:t ] = z[ ll:rr, bb:tt ]
    return w

def insert(im,a,p,action='add'):
    """insert a small image im into a larger array a at position p.
    im: small image to insert
    a: bigger image into which to insert the smaller image
    p: center position in a at which to insert im's center
    action can be 'add' or 'replace'
    """
    n,m = im.shape
    na,ma = a.shape
    pll = [p[0]-n/2,p[1]-m/2]
    pur = [pll[0]+n,pll[1]+m]
    qll = [0,0]
    qur = [n,m]
    for k,lim in zip([0,1],[na,ma]):
        if pll[k] < 0:
            d = abs(pll[k])
            pll[k] += d
            qll[k] += d
        if pur[k] >= lim:
            d = pur[k]-lim + 1
            pur[k] -= d
            qur[k] -= d
        if pll[k] >= pur[k]:
            return a
    if action == 'add':
        a[pll[0]:pur[0],pll[1]:pur[1]] += im[qll[0]:qur[0],qll[1]:qur[1]]
    elif action == 'replace':
        a[pll[0]:pur[0],pll[1]:pur[1]] = im[qll[0]:qur[0],qll[1]:qur[1]]

    return a

def rebin(a, shape):
    '''rebin ndarray data into a smaller ndarray of the same rank whose dimensions
    are factors of the original dimensions. eg. An array with 6 columns and 4 rows
    can be reduced to have 6,3,2 or 1 columns and 4,2 or 1 rows.
    example usages:
    >>> a=rand(6,4); b=rebin(a,3,2)
    >>> a=rand(6); b=rebin(a,2)

    Reference: http://wiki.scipy.org/Cookbook/Rebinning
    http://stackoverflow.com/questions/8090229/resize-with-averaging-or-rebin-a-numpy-2d-array
    '''
    sh = shape[0],a.shape[0]//shape[0], shape[1],a.shape[1]//shape[1]
    return a.reshape(sh).mean(-1).mean(1)

global newcoords
global newcoords_dims

def congrid(a, newdims, method='linear', centre=False, minusone=False):
    '''Arbitrary resampling of source array to new dimension sizes.
    Currently only supports maintaining the same number of dimensions.
    To use 1-D arrays, first promote them to shape (x,1).
    
    Uses the same parameters and creates the same co-ordinate lookup points
    as IDL''s congrid routine, which apparently originally came from a VAX/VMS
    routine of the same name.

    method:
    neighbour - closest value from original data
    nearest and linear - uses n x 1-D interpolations using
                         scipy.interpolate.interp1d
    (see Numerical Recipes for validity of use of n 1-D interpolations)
    spline - uses ndimage.map_coordinates

    centre:
    True - interpolation points are at the centres of the bins
    False - points are at the front edge of the bin

    minusone:
    For example- inarray.shape = (i,j) & new dimensions = (x,y)
    False - inarray is resampled by factors of (i/x) * (j/y)
    True - inarray is resampled by(i-1)/(x-1) * (j-1)/(y-1)
    This prevents extrapolation one element beyond bounds of input array.
    
    Reference: http://wiki.scipy.org/Cookbook/Rebinning
    '''
    global newcoords
    global newcoords_dims
    if not a.dtype in [np.float64, np.float32]:
        a = np.cast[float](a)

    m1 = np.cast[int](minusone)
    ofs = np.cast[int](centre) * 0.5
    old = np.array( a.shape )
    ndims = len( a.shape )
    if len( newdims ) != ndims:
        print "[congrid] dimensions error. " \
              "This routine currently only support " \
              "rebinning to the same number of dimensions."
        return None
    newdims = np.asarray( newdims, dtype=float )
    dimlist = []

    if method == 'neighbour':
        for i in range( ndims ):
            base = np.indices(newdims)[i]
            dimlist.append( (old[i] - m1) / (newdims[i] - m1) \
                            * (base + ofs) - ofs )
        cd = np.array( dimlist ).round().astype(int)
        newa = a[list( cd )]
        return newa

    elif method in ['nearest','linear']:
        # calculate new dims
        for i in range( ndims ):
            base = np.arange( newdims[i] )
            dimlist.append( (old[i] - m1) / (newdims[i] - m1) \
                            * (base + ofs) - ofs )
        # specify old dims
        olddims = [np.arange(i, dtype = np.float) for i in list( a.shape )]

        # first interpolation - for ndims = any
        mint = scipy.interpolate.interp1d( olddims[-1], a, kind=method )
        newa = mint( dimlist[-1] )

        trorder = [ndims - 1] + range( ndims - 1 )
        for i in range( ndims - 2, -1, -1 ):
            newa = newa.transpose( trorder )

            mint = scipy.interpolate.interp1d( olddims[i], newa, kind=method )
            newa = mint( dimlist[i] )

        if ndims > 1:
            # need one more transpose to return to original dimensions
            newa = newa.transpose( trorder )

        return newa
    elif method in ['spline']:
        oslices = [ slice(0,j) for j in old ]
        oldcoords = np.ogrid[oslices]
        nslices = [ slice(0,j) for j in list(newdims) ]
        newcoords = np.mgrid[nslices]

        newcoords_dims = range(np.ndim(newcoords))
        #make first index last
        newcoords_dims.append(newcoords_dims.pop(0))
        newcoords_tr = newcoords.transpose(newcoords_dims)
        # makes a view that affects newcoords

        newcoords_tr += ofs

        deltas = (np.asarray(old) - m1) / (newdims - m1)
        newcoords_tr *= deltas

        newcoords_tr -= ofs

        newa = scipy.ndimage.map_coordinates(a, newcoords)
        return newa
    else:
        print "Congrid error: Unrecognized interpolation type.\n", \
              "Currently only \'neighbour\', \'nearest\',\'linear\',", \
              "and \'spline\' are supported."
        return None
    
def minmax(a):
    """Return the minimum and maximum of an array as a 2-vector
    """
    return a.min(),a.max()

def centroid(z,ap=None):
    """Find the centroid of data
    
    inputs:
        z - a 2-D array
    
    keyword arguments:
        ap - an aperture over which to compute the centroid
    
    returns:
        the centroid, in pixel units, with 0,0 defined as the lower-left of the 0,0 pixel
        (pixel k spans the continuum range from k to k+1)
    """
    n,m = z.shape
    x = np.arange(n)+0.5 # half-pixel shift. e.g. 0,0 pixel = 1 centroids to position 0.5,0.5
    x,y = np.meshgrid(x,x)
    if (ap is None):
        ap = np.ones((n,m))
    zap = z*ap
    zap /= np.sum(zap)
    cx = np.sum(x*zap)
    cy = np.sum(y*zap)
    return cx,cy

def azimuthalAverage(image, center=None, stddev=False, returnradii=False, return_nr=False, 
        binsize=3., weights=None, steps=False, interpnan=False, left=None, right=None):
    """
    https://code.google.com/p/agpy/source/browse/trunk/agpy/radialprofile.py?r=317
    
    Calculate the azimuthally averaged radial profile.

    image - The 2D image
    center - The [x,y] pixel coordinates used as the center. The default is 
             None, which then uses the center of the image (including 
             fractional pixels).
    stddev - if specified, return the azimuthal standard deviation instead of the average
    returnradii - if specified, return (radii_array,radial_profile)
    return_nr   - if specified, return number of pixels per radius *and* radius
    binsize - size of the averaging bin.  Can lead to strange results if
        non-binsize factors are used to specify the center and the binsize is
        too large
    weights - can do a weighted average instead of a simple average if this keyword parameter
        is set.  weights.shape must = image.shape.  weighted stddev is undefined, so don't
        set weights and stddev.
    steps - if specified, will return a double-length bin array and radial
        profile so you can plot a step-form radial profile (which more accurately
        represents what's going on)
    interpnan - Interpolate over NAN values, i.e. bins where there is no data?
        left,right - passed to interpnan; they set the extrapolated values

    If a bin contains NO DATA, it will have a NAN value because of the
    divide-by-sum-of-weights component.  I think this is a useful way to denote
    lack of data, but users let me know if an alternative is prefered...
    
    """
    # Calculate the indices from the image
    n,m = image.shape
    y, x = np.indices(image.shape)

    if center is None:
        center = np.array([float(n/2),float(m/2)])
        #center = np.array([(x.max()-x.min())/2.0, (y.max()-y.min())/2.0])

    r = np.hypot(x - center[0], y - center[1])

    if weights is None:
        weights = np.ones(image.shape)
    elif stddev:
        raise ValueError("Weighted standard deviation is not defined.")

    # the 'bins' as initially defined are lower/upper bounds for each bin
    # so that values will be in [lower,upper)  
    nbins = (np.round(r.max() / binsize)+1)
    maxbin = nbins * binsize
    bins = np.linspace(0,maxbin,nbins+1)
    # but we're probably more interested in the bin centers than their left or right sides...
    bin_centers = (bins[1:]+bins[:-1])/2.0
    
    # Find out which radial bin each point in the map belongs to
    whichbin = np.digitize(r.flat,bins)
    
    # how many per bin (i.e., histogram)?
    # there are never any in bin 0, because the lowest index returned by digitize is 1
    nr = np.bincount(whichbin)[1:]

    # recall that bins are from 1 to nbins (which is expressed in array terms by arange(nbins)+1 or xrange(1,nbins+1) )
    # radial_prof.shape = bin_centers.shape
    if stddev:
        radial_prof = np.array([image.flat[whichbin==b].std() for b in (np.arange(nbins)+1)]) #xrange(1,nbins+1)])
    else:
        radial_prof = np.array([(image*weights).flat[whichbin==b].sum() / weights.flat[whichbin==b].sum() for b in (np.arange(nbins)+1)]) #xrange(1,nbins+1)])

    #import pdb; pdb.set_trace()

    if interpnan:
        radial_prof = np.interp(bin_centers,bin_centers[radial_prof==radial_prof],radial_prof[radial_prof==radial_prof],left=left,right=right)

    if steps:
        xarr = np.array(zip(bins[:-1],bins[1:])).ravel() 
        yarr = np.array(zip(radial_prof,radial_prof)).ravel() 
        return xarr,yarr
    elif returnradii: 
        return bin_centers,radial_prof
    elif return_nr:
        return nr,bin_centers,radial_prof
    else:
        return radial_prof

def zoom(a,zoom_factor,shape=None,order=3,center=None):
    """Zoom the image by resampling to a different pixel scale.
    The routine wraps the scipy.ndimage.interpolation.zoom function
    but also allows for specifying the shape of the result image,
    which is either cropped or zeropadded to fit.
    
    Argument:
        a = the image
        zoom_factor = the ratio of the old dx to the new dx,
                      i.e. a 2:1 zoom takes old dx=1 to new dx=0.5
    
    Keyword:
        shape = None: return a reshaped array that fits the data (default)
                'same': returns an array the same shape as a
                2-tuple: a desired shape
    
    """
    n,m = a.shape
    nr,mr = [int(float(x)*zoom_factor) for x in a.shape] # this is the resulting shape
    if shape is None:
        shape = nr,mr
    elif shape is 'same':
        shape = n,m

    if center is None:
        center = (n/2,m/2)
    else: # imbed the image in a zero-padded enclosing rectangle centered on the new center
        nb,mb = [ 2*max(c,(nn-c)) for (nn,c) in zip(a.shape,center) ]
        x,y = [ max(0,nn-2*c) for (nn,c) in zip(a.shape,center) ]
        aa = np.zeros((nb,mb))
        aa[x:x+n,y:y+m] = a
        a = aa
        n,m = a.shape
    
    if shape[0] < nr or shape[1] < mr :  #zooming in to a desired shape that is smaller than the resulting shape
        subshape = [int(float(x)/zoom_factor) for x in shape]
        a = crop(a,(n/2,m/2),subshape)
    r = scipy.ndimage.zoom(a,zoom_factor,order=order)
    return r

# -------------------- Tests ---------------------------
def test():
    """Test the airy and psf routines
    """
    n = 256
    dp = 3.
    ds = 0.8
    lam = 1.6e-6
    pixelscale = 0.033*units.arcsec
    s = 1.219669891266504*(lam/dp)/pixelscale
    print '<test> computing a numerical PSF'
    u = psf(n,dp,0,lam,pixelscale,pad=4)
    u = u/u.max()
    print '<test> computing an analytic PSF'
    v = airy(n,n,s)
    print '<text> max difference: '+str(np.abs(u-v).max()) + ' (compare to 1.0 peak)'
    
def test2():
    """Test of crop and insert routines
    """
    global z,zc,zc2
    n = 4
    z = np.zeros((n,n))
    z[1:3,1:3] = 1.
    print z
    zc = crop(z,(2,2),(2,2))
    print zc
    a = np.zeros((n,n))
    a = insert(zc,a,(2,2),action='replace')
    print a
    #
    zc = np.ones((33,33))
    z = np.zeros((200,200))
    z = insert(zc,z,(100,100),action='replace')
    zc2 = crop(z,(100,100),(33,33))
    



from setuptools import setup, find_packages
execfile('glao_psf/_version.py')

setup(
  name = 'glao_psf',
  packages = find_packages(),
  version = __version__,
  description = 'Adaptive Optics Structure Function and Average PSF calculator',
  author = 'Don Gavel',
  author_email = 'donald.gavel@gmail.com',
  url = 'https://bitbucket.org/donald_gavel/glao_psf', # the github repo
  install_requires=[
    'matplotlib',
    'six',
    'numpy',
    'scipy',
    'pyfits',
    'h5py',
    'tables',
    'pandas',
  ],
  include_package_data = True,
  download_url = 'https://bitbucket.org/donald_gavel/glao_psf/downloads',
  keywords = ['Adaptive Optics'], # arbitrary keywords
  classifiers = [],
)
# 



"""
pprint is an enhanced pretty-print for objects
"""
import numpy as np
import pprint as up_pprint
import json
from collections import OrderedDict

def ppd(self,print_=True):
    """Pretty-print the dictionary
    """
    if isinstance(self,dict):
        try:
            r = json.dumps(self,indent=4)
        except:
            d = OrderedDict()
            valid_types = (int,long,float,complex,str)
            for key,val in zip(self.keys(),self.values()):
                if isinstance(val,valid_types):
                    d[key] = val
                else:
                    d[key] = str(type(val))
            r = json.dumps(d,indent=4)
    else:
        r = ''
    if print_:
        print r
    else:
        return r

def pprint(self,print_width=60, short=False, expand=0):
    """Pretty-print the object. This consists
    of reporting on the object's attributes.
    
    :param object self: an object of any type
    :param int print_width: widest printout width per attribute.
        After this the output is abreviated
    :param boolean short: short output. Just the name
        and object type, no attributes.
    "param boolean expand: expand all dictionaries in a nested indented form
        level 1 does only dicts, level 2 does lists of dicts

    """
    if isinstance(self,dict):
        ppd(self)
        return
    elif isinstance(self,(list,tuple)):
        for e in self:
            pprint(e,print_width=print_width,short=short,expand=expand)
        return
    elif not hasattr(self,'__dict__'):
        up_pprint.pprint(self)
        return
    bling = print_width*'='
    print bling
    print type(self),
    if hasattr(self,'longName'):
        print self.longName
    elif hasattr(self,'name'):
        print self.name
    else:
        print ''
    print print_width*'-'
    if short:
        return
    tab = ' '*4
    d = self.__dict__
    keys = sorted(d.keys())
    for key in keys:
        limit_length = True
        if key.endswith('_units'): continue
        val = d[key]
        units = None
        
        if isinstance(val,(str,unicode)):
            pv = "'"+val+"'"
        elif isinstance(val,(tuple,list)):
            if (len(val) < 5) and (np.array(map(np.isscalar,val)).all()):
                pv = str(val)
            elif expand > 1:
                pv = str(type(val)) + ' ' + str(len(val)) + ' entries'
                for e in val:
                    r = ppd(e,print_=False)
                    if r == '':
                        r = str(type(e))
                    else:
                        limit_length = False
                    pv += '\n' + r
            else:
                pv = str(type(val)) + ' ' + str(len(val)) + ' entries'
        elif expand > 0 and isinstance(val,dict):
            pv = ppd(val,print_=False)
            limit_length = False
        else:
            pv = str(val)
            
        if limit_length:
            if key.endswith('_doc'):
                pass
            elif len(pv) > print_width:
                pv = str(val.__class__)
                if isinstance(val,(np.ndarray,np.matrix)):
                    pv += ' '+str(val.shape)
                
        if key+'_units' in keys:
            units = d[key+'_units']
        line = key+':'+tab+pv
        if units is None:
            pass
        else:
            line += ' '+units
        if isinstance(val,object):
            if (hasattr(val,'longName')):
                line += " '"+val.longName+"'"
            elif (hasattr(val,'name')):
                line += " '"+val.name+"'"
        else:
            pass
        print line
    print bling

def methods(self):
    cm = dir(self.__class__)
    return [x for x in cm if not x.startswith('__')]

def test():
    from info_array import InfoArray
    pprint(1)
    pprint('hello')
    a = InfoArray([[1,2],[3,4]],name='foo')
    pprint(a)
    d = OrderedDict()
    d['one'] = 'hello'
    d['two'] = 12.34
    d['three'] = 24
    d['four'] = np.array([1,2])
    pprint(d)


"""
Common units.
Defines a set of commonly used units for use
in physical modeling and display

Typical usage::

    >>> from common_units import *
    
Then for example, access the 'micron' unit conversion with::

    >>> units['micron']
    1.e-06
    >>> units.micron
    1.e-06
    >>> micron
    1.e-06
    
which are all= 1.e-6

The import * adds a lot of variable names to the local
environment, but these are only the unit names
such as 'mm', 'cm', etc. (those listed by
units.list_all()) and the name 'units'.
If you don't want this, you can::

    >>> from common_units import units
    
Then::

    >>> units['micron'] # <-- works okay
    1e-06
    >>> units.micron  # <-- works okay
    1e-06
    >>> micron  # <-- doesnt work
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
    NameError: name 'micron' is not defined

You can always load them into the local environment later with a call to
:py:meth:`put_into <common_units.Units.put_into>`.
"""
import sys
from numpy import pi

class Units(object):
    """The Units class defines an object with
    unit conversion factors
    """
    def __init__(self):
        self.arcsec = pi/(180.*3600.)
        self.arcmin = pi/(180.*60.)
        self.meters = 1.
        self.degree = pi/180.
        self.degrees = self.degree
        self.centimeters = 1.e-2
        self.centimeter = self.centimeters
        self.cm = self.centimeter
        self.millimeters = 1.e-3
        self.millimeter = self.millimeters
        self.mm = self.millimeters
        self.microns = 1.e-6
        self.micron = self.microns
        self.nanometers = 1.e-9
        self.nanometer = self.nanometers
        self.nm = self.nanometers
        self.kilometers = 1.e3
        self.kilometer = self.kilometers
        self.km = self.kilometers
        self.second = 1.
        self.seconds = self.second
        self.millisecond = 0.001
        self.milliseconds = self.millisecond
        self.microsecond = 1.e-6
        self.microseconds = self.microsecond
        self.ms = self.millisecond
        self._load()
        
    def list_all(self):
        """List all the unit conversion factors
        and their values.
        
        The current set of units is:
        
            >>> units.list_all()
            arcsec 4.8481368111e-06
            centimeter 0.01
            centimeters 0.01
            cm 0.01
            kilometer 1000.0
            kilometers 1000.0
            meters 1.0
            micron 1e-06
            microns 1e-06
            microsecond 1e-06
            microseconds 1e-06
            millimeter 0.001
            millimeters 0.001
            millisecond 0.001
            milliseconds 0.001
            mm 0.001
            ms 0.001
            nanometer 1e-09
            nanometers 1e-09
            nm 1e-09
            second 1.0
            seconds 1.0
        
        """
        keys = sorted(self.__dict__.keys())
        for key in keys:
            print key,self[key]
    
    def get_dict(self):
        """returns a dictionary with all the
        defined unit conversion factors
        """
        return self.__dict__
    
    def keys(self):
        """returns the list of unit names
        """
        return self.__dict__.keys()
    
    def __getitem__(self,key):
        return self.__dict__[key]
    
    def _load(self,module=None):
        if module is None:
            module = sys.modules[__name__]
        elif isinstance(module,str):
            module = sys.modules[module]
        for name,val in self.get_dict().iteritems():
            setattr(module,name,val)
        setattr(module,'units',self)
    
    def put_into(self,local_dict):
        """update a dictionary with the units.
        For example, to import them into the top-level (globals)::
        
            >>> from common_units import units
            >>> units.put_into(globals())
            
        """
        local_dict.update(self.get_dict())
        local_dict.update({'units':units})

units = Units()

def test():
    print units.get_dict()
    print units['mm']
    print units.mm
    print mm
